#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    In information retrieval, tf–idf, TF*IDF, or TFIDF, short for term frequency–inverse document frequency,
    is a numerical statistic that is intended to reflect how important a word is to a document in a collection or corpus.
    It is often used as a weighting factor in searches of information retrieval, text mining, and user modeling.

    The tf–idf value increases proportionally to the number of times a word appears in the document
    and is offset by the number of documents in the corpus that contain the word,
    which helps to adjust for the fact that some words appear more frequently in general.

    tf–idf is one of the most popular term-weighting schemes today.
    A survey conducted in 2015 showed that 83% of text-based recommender systems in digital libraries use tf–idf.

    Variations of the tf–idf weighting scheme are often used by search engines as a central tool
    in scoring and ranking a document's relevance given a user query.
    tf–idf can be successfully used for stop-words filtering in various subject fields,
    including text summarization and classification.

    One of the simplest ranking functions is computed by summing the tf–idf for each query term;
    many more sophisticated ranking functions are variants of this simple model.
"""
import re
import typing
from math import log

from borb.pdf.page.page import Page
from borb.toolkit.text.simple_text_extraction import SimpleTextExtraction


class TFIDFKeywordExtraction(SimpleTextExtraction):
    """
    In information retrieval, tf–idf, TF*IDF, or TFIDF, short for term frequency–inverse document frequency,
    is a numerical statistic that is intended to reflect how important a word is to a document in a collection or corpus.
    It is often used as a weighting factor in searches of information retrieval, text mining, and user modeling.

    The tf–idf value increases proportionally to the number of times a word appears in the document
    and is offset by the number of documents in the corpus that contain the word,
    which helps to adjust for the fact that some words appear more frequently in general.

    tf–idf is one of the most popular term-weighting schemes today.
    A survey conducted in 2015 showed that 83% of text-based recommender systems in digital libraries use tf–idf.

    Variations of the tf–idf weighting scheme are often used by search engines as a central tool
    in scoring and ranking a document's relevance given a user query.
    tf–idf can be successfully used for stop-words filtering in various subject fields,
    including text summarization and classification.

    One of the simplest ranking functions is computed by summing the tf–idf for each query term;
    many more sophisticated ranking functions are variants of this simple model.
    """

    def __init__(
        self, stopwords: typing.List[str] = [], minimum_term_frequency: int = 3
    ):
        super().__init__()
        self._term_frequency_per_page: typing.Dict[int, typing.Dict[str, float]] = {}
        self._number_of_words_per_page: typing.Dict[int, float] = {}
        self._inverse_page_frequency: typing.Dict[str, float] = {}
        self._stopwords: typing.List[str] = [x.upper() for x in stopwords]
        self._number_of_pages: float = 0
        self._minimum_term_frequency: float = minimum_term_frequency

    def _end_page(self, page: Page):
        super()._end_page(page)

        # update number of pages
        self._number_of_pages += 1

        # get words
        words_on_page: typing.List[str] = [
            x.upper()
            for x in re.split("[^a-zA-Z]+", self.get_text_for_page(self._current_page))
            if x.upper() not in self._stopwords
        ]

        # update _term_frequency_per_page
        self._term_frequency_per_page[self._current_page] = {}
        self._number_of_words_per_page[self._current_page] = len(words_on_page)
        for w in words_on_page:
            self._term_frequency_per_page[self._current_page][w] = (
                self._term_frequency_per_page[self._current_page].get(w, 0) + 1
            )

        # update _inverse_page_frequency
        for w in set(words_on_page):
            self._inverse_page_frequency[w] = self._inverse_page_frequency.get(w, 0) + 1

    def get_keywords_for_page(
        self, page_number: int
    ) -> typing.List[typing.Tuple[str, float]]:
        """
        This function returns a typing.List[typing.Tuple[str, float]] for a given page
        """
        out: typing.List[typing.Tuple[str, float]] = []
        for w, tf in self._term_frequency_per_page[page_number].items():

            # check minimum_term_frequency
            if tf < self._minimum_term_frequency:
                continue

            # normalize tf, idf
            tf /= self._number_of_words_per_page[page_number]
            idf = log(self._number_of_pages / self._inverse_page_frequency[w])

            # avoid multiply by zero
            tf += 0.0001
            idf += 0.0001

            # calculate tf-idf score
            out.append((w, tf * idf))

        # sort
        out.sort(key=lambda x: x[1], reverse=True)

        # return
        return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Move to the start of the next line. This operator has the same effect as the
code
0 -Tl Td
where Tl denotes the current leading parameter in the text state. The
negative of Tl is used here because Tl is the text leading expressed as a
positive number. Going to the next line entails decreasing the
y coordinate.
"""
import typing

from borb.io.read.types import AnyPDFType
from borb.io.read.types import Decimal as bDecimal
from borb.pdf.canvas.operator.canvas_operator import CanvasOperator


class MoveToNextLine(CanvasOperator):
    """
    Move to the start of the next line. This operator has the same effect as the
    code
    0 -Tl Td
    where Tl denotes the current leading parameter in the text state. The
    negative of Tl is used here because Tl is the text leading expressed as a
    positive number. Going to the next line entails decreasing the
    y coordinate.
    """

    def __init__(self):
        super().__init__("T*", 0)

    def invoke(
        self,
        canvas_stream_processor: "CanvasStreamProcessor",  # type: ignore [name-defined]
        operands: typing.List[AnyPDFType] = [],
        event_listeners: typing.List["EventListener"] = [],  # type: ignore [name-defined]
    ) -> None:
        """
        Invoke the T* operator
        """
        move_text_position_op: typing.Optional[
            CanvasOperator
        ] = canvas_stream_processor.get_operator("Td")
        assert (
            move_text_position_op
        ), "Operator Td must be defined for operator T* to function."
        canvas = canvas_stream_processor.get_canvas()
        move_text_position_op.invoke(
            canvas_stream_processor,
            [bDecimal(0), -canvas.graphics_state.leading],
            event_listeners,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This implementation of FormField represents a push button.
"""
import typing
import zlib
from decimal import Decimal

from borb.io.read.types import Decimal as bDecimal
from borb.io.read.types import Dictionary, Name, List, String, Stream, Boolean
from borb.pdf.canvas.color.color import HexColor, Color
from borb.pdf.canvas.font.simple_font.font_type_1 import StandardType1Font
from borb.pdf.canvas.geometry.rectangle import Rectangle
from borb.pdf.canvas.layout.forms.form_field import FormField
from borb.pdf.canvas.layout.layout_element import Alignment
from borb.pdf.canvas.layout.text.line_of_text import LineOfText
from borb.pdf.page.page import Page


class PushButton(FormField):
    """
    This implementation of FormField represents a push button.
    """

    def __init__(
        self,
        text: str,
        background_color: typing.Optional[Color] = HexColor("efefef"),
        border_bottom: bool = True,
        border_color: Color = HexColor("767676"),
        border_left: bool = True,
        border_right: bool = True,
        border_top: bool = True,
        border_width: Decimal = Decimal(1),
        field_name: typing.Optional[str] = None,
        font_size: Decimal = Decimal(12),
        font_color: Color = HexColor("000000"),
        horizontal_alignment: Alignment = Alignment.LEFT,
        margin_bottom: typing.Optional[Decimal] = Decimal(0),
        margin_left: typing.Optional[Decimal] = Decimal(0),
        margin_right: typing.Optional[Decimal] = Decimal(0),
        margin_top: typing.Optional[Decimal] = Decimal(0),
        padding_bottom: Decimal = Decimal(2),
        padding_left: Decimal = Decimal(6),
        padding_right: Decimal = Decimal(6),
        padding_top: Decimal = Decimal(2),
    ):
        super(PushButton, self).__init__()
        assert len(text) > 0
        self._text = text
        self._background_color = background_color
        self._border_bottom = border_bottom
        self._border_color = border_color
        self._border_left = border_left
        self._border_right = border_right
        self._border_top = border_top
        assert border_width >= 0
        self._border_width = border_width
        self._field_name: typing.Optional[str] = field_name
        assert font_size > 0
        self._font_size = font_size
        self._font_color = font_color
        self._horizontal_alignment = horizontal_alignment
        assert margin_bottom is not None and margin_bottom >= 0
        self._margin_bottom = margin_bottom
        assert margin_left is not None and margin_left >= 0
        self._margin_left = margin_left
        assert margin_right is not None and margin_right >= 0
        self._margin_right = margin_right
        assert margin_top is not None and margin_top >= 0
        self._margin_top = margin_top
        assert padding_bottom >= 0
        self._padding_bottom = padding_bottom
        assert padding_left >= 0
        self._padding_left = padding_left
        assert padding_right >= 0
        self._padding_right = padding_right
        assert padding_top >= 0
        self._padding_top = padding_top
        self._widget_dictionary: typing.Optional[Dictionary] = None

    def _init_widget_dictionary(self, page: Page) -> None:
        if self._widget_dictionary is not None:
            return

        # init page and font resources
        assert self._font_size is not None
        font_resource_name: Name = self._get_font_resource_name(
            StandardType1Font("Helvetica"), page
        )

        # widget resource dictionary
        widget_resources: Dictionary = Dictionary().set_is_unique(True)  # type: ignore [attr-defined]
        widget_resources[Name("Font")] = page["Resources"]["Font"]

        # get Catalog
        catalog: Dictionary = page.get_root()["XRef"]["Trailer"]["Root"]  # type: ignore [attr-defined]

        # widget dictionary
        # fmt: off
        self._widget_dictionary = Dictionary().set_is_unique(True)     # type: ignore [attr-defined]
        self._widget_dictionary.set_is_unique(True) # type: ignore [attr-defined]
        self._widget_dictionary[Name("AA")] = Dictionary()
        self._widget_dictionary[Name("AA")][Name("D")] = Dictionary()
        self._widget_dictionary[Name("AA")][Name("D")][Name("Type")] = Name("Action")
        self._widget_dictionary[Name("AA")][Name("D")][Name("S")] = Name("ResetForm")
        # fmt: on

        # create normal appearance
        # fmt: off
        self._widget_dictionary[Name("AP")] = Dictionary().set_is_unique(True)     # type: ignore [attr-defined]
        self._widget_dictionary[Name("AP")][Name("N")] = Stream()
        self._widget_dictionary[Name("AP")][Name("N")][Name("Type")] = Name("XObject")
        self._widget_dictionary[Name("AP")][Name("N")][Name("Subtype")] = Name("Form")
        self._widget_dictionary[Name("AP")][Name("N")][Name("BBox")] = List().set_is_inline(True)   # type: ignore[attr-defined]
        for _ in range(0, 4):
            self._widget_dictionary[Name("AP")][Name("N")][Name("BBox")].append(bDecimal(0))
        self._widget_dictionary[Name("AP")][Name("N")][Name("DecodedBytes")] = b"/Tx BMC EMC"
        self._widget_dictionary[Name("AP")][Name("N")][Name("Bytes")] = zlib.compress(self._widget_dictionary[Name("AP")][Name("N")][Name("DecodedBytes")], 9)
        self._widget_dictionary[Name("AP")][Name("N")][Name("Filter")] = Name("FlateDecode")
        self._widget_dictionary[Name("AP")][Name("N")][Name("Length")] = bDecimal(len(self._widget_dictionary[Name("AP")][Name("N")][Name("Bytes")]))
        self._widget_dictionary[Name("AP")][Name("N")][Name("Resources")] = Dictionary()
        self._widget_dictionary[Name("AP")][Name("N")][Name("Resources")][Name("ProcSet")] = List().set_is_inline(True) # type: ignore [attr-defined]
        self._widget_dictionary[Name("AP")][Name("N")][Name("Resources")][Name("ProcSet")].append(Name("PDF"))
        self._widget_dictionary[Name("AP")][Name("N")][Name("Resources")][Name("ProcSet")].append(Name("Text"))
        self._widget_dictionary[Name("AP")][Name("N")][Name("Resources")][Name("Font")] = Dictionary()
        self._widget_dictionary[Name("AP")][Name("N")][Name("Resources")][Name("Font")] = page["Resources"]["Font"]
        # fmt: on

        # create default appearance
        # fmt: off
        self._widget_dictionary[Name("DA")] = String("0.23921 0.23921 0.23921 rg /%s %f Tf" % (font_resource_name, float(self._font_size)))
        self._widget_dictionary[Name("DR")] = widget_resources
        # fmt: on

        # other flags
        # fmt: off
        self._widget_dictionary[Name("F")] = bDecimal(4)
        self._widget_dictionary[Name("Ff")] = bDecimal(65536)
        self._widget_dictionary[Name("FT")] = Name("Btn")
        self._widget_dictionary[Name("MK")] = Dictionary()
        self._widget_dictionary[Name("MK")][Name("BC")] = List().set_is_inline(True)    # type: ignore [attr-defined]
        self._widget_dictionary[Name("MK")][Name("BG")] = List().set_is_inline(True)    # type: ignore [attr-defined]
        self._widget_dictionary[Name("MK")][Name("CA")] = String("")
        self._widget_dictionary[Name("P")] = catalog
        self._widget_dictionary[Name("Q")] = bDecimal(1)
        self._widget_dictionary[Name("Rect")] = List().set_is_inline(True)              # type: ignore [attr-defined]
        for _ in range(0, 4):
            self._widget_dictionary[Name("Rect")].append(bDecimal(0))
        self._widget_dictionary[Name("Subtype")] = Name("Widget")
        self._widget_dictionary[Name("T")] = String(self._field_name or self._get_auto_generated_field_name(page))
        self._widget_dictionary[Name("Type")] = Name("Annot")
        # fmt: on

        # append field to page /Annots
        if "Annots" not in page:
            page[Name("Annots")] = List()
        page["Annots"].append(self._widget_dictionary)

        # append field to catalog
        if "AcroForm" not in catalog:
            catalog[Name("AcroForm")] = Dictionary()
            catalog["AcroForm"][Name("Fields")] = List()
            catalog["AcroForm"][Name("DR")] = widget_resources
            catalog["AcroForm"][Name("NeedAppearances")] = Boolean(True)
        catalog["AcroForm"]["Fields"].append(self._widget_dictionary)

    def _get_content_box(self, available_space: Rectangle) -> Rectangle:
        assert self._font_size is not None
        line_of_text: LineOfText = LineOfText(
            self._text,
            background_color=self._background_color,
            font_size=self._font_size,
            font_color=self._font_color,
        )
        return line_of_text.get_layout_box(available_space)

    def _paint_content_box(self, page: "Page", available_space: Rectangle) -> None:

        # init self._widget_dictionary
        self._init_widget_dictionary(page)

        # layout text
        assert self._font_size is not None
        line_of_text = LineOfText(
            self._text,
            background_color=self._background_color,
            font_size=self._font_size,
            font_color=self._font_color,
        )
        line_of_text.paint(page, self._get_content_box(available_space))
        cbox: typing.Optional[Rectangle] = line_of_text.get_previous_paint_box()
        assert cbox is not None

        # add our own margins to expand the clickable part
        cbox.x -= self._padding_left
        cbox.width += self._padding_left + self._padding_right
        cbox.y -= self._padding_bottom
        cbox.height += self._padding_bottom + self._padding_top

        # set location
        # fmt: off
        if self._widget_dictionary is not None:
            self._widget_dictionary["Rect"][0] = bDecimal(cbox.get_x())                             # ll_x
            self._widget_dictionary["Rect"][1] = bDecimal(cbox.get_y())                             # ll_y
            self._widget_dictionary["Rect"][2] = bDecimal(cbox.get_x() + cbox.get_width())          # ur_x
            self._widget_dictionary["Rect"][3] = bDecimal(cbox.get_y() + cbox.get_height())         # ur_y
        # fmt: on


class JavaScriptPushButton(PushButton):
    """
    This implementation of FormField represents a push button that triggers JavaScript.
    """

    def __init__(
        self,
        javascript: str,
        text: str,
        background_color: typing.Optional[Color] = HexColor("efefef"),
        border_bottom: bool = True,
        border_color: Color = HexColor("767676"),
        border_left: bool = True,
        border_right: bool = True,
        border_top: bool = True,
        border_width: Decimal = Decimal(1),
        field_name: typing.Optional[str] = None,
        font_size: Decimal = Decimal(12),
        font_color: Color = HexColor("000000"),
        horizontal_alignment: Alignment = Alignment.LEFT,
        margin_bottom: Decimal = Decimal(0),
        margin_left: Decimal = Decimal(0),
        margin_right: Decimal = Decimal(0),
        margin_top: Decimal = Decimal(0),
        padding_bottom: Decimal = Decimal(2),
        padding_left: Decimal = Decimal(6),
        padding_right: Decimal = Decimal(6),
        padding_top: Decimal = Decimal(2),
    ):
        super(JavaScriptPushButton, self).__init__(
            text=text,
            background_color=background_color,
            border_bottom=border_bottom,
            border_color=border_color,
            border_left=border_left,
            border_right=border_right,
            border_top=border_top,
            border_width=border_width,
            field_name=field_name,
            font_size=font_size,
            font_color=font_color,
            horizontal_alignment=horizontal_alignment,
            margin_bottom=margin_bottom,
            margin_left=margin_left,
            margin_right=margin_right,
            margin_top=margin_top,
            padding_bottom=padding_bottom,
            padding_left=padding_left,
            padding_right=padding_right,
            padding_top=padding_top,
        )
        self._javascript: str = javascript

    def _init_widget_dictionary(self, page: Page) -> None:

        # call to super
        super(JavaScriptPushButton, self)._init_widget_dictionary(page)

        # build JavaScript stream object
        # fmt: off
        javascript_stream = Stream().set_is_unique(True)     # type: ignore [attr-defined]
        javascript_stream[Name("Type")] = Name("JavaScript")
        javascript_stream[Name("DecodedBytes")] = bytes(self._javascript, "latin1")
        javascript_stream[Name("Bytes")] = zlib.compress(javascript_stream[Name("DecodedBytes")], 9)
        javascript_stream[Name("Length")] = bDecimal(len(javascript_stream[Name("Bytes")]))
        javascript_stream[Name("Filter")] = Name("FlateDecode")
        # fmt: on

        # modify action dictionary of PushButton (super)
        # fmt: off
        if self._widget_dictionary is not None:
            self._widget_dictionary[Name("AA")][Name("D")][Name("S")] = Name("JavaScript")
            self._widget_dictionary[Name("AA")][Name("D")][Name("JS")] = javascript_stream
        # fmt: on



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    This module defines an annotation used for profiling
"""
import cProfile
import io
import pstats
from pstats import SortKey  # type: ignore [attr-defined]


def profile(func):
    """
    This annotation wraps itself around a function, profiling it
    """

    def wrapper(*args, **kwargs):
        """
        This function starts the profiling, executes the function (storing the result), outputs the profiling information
        and returns the result
        """
        # setup profiler
        pr = cProfile.Profile()
        pr.enable()

        # execute function
        out_value = func(*args, **kwargs)

        # disable profiler
        pr.disable()
        s = io.StringIO()
        sortby = SortKey.CUMULATIVE
        ps = pstats.Stats(pr, stream=s).sort_stats(sortby)
        ps.print_stats()
        print(s.getvalue())

        # return
        return out_value

    return wrapper



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
A printer’s mark annotation (PDF 1.4) represents a graphic symbol, such as a registration target, colour bar, or
cut mark, that may be added to a page to assist production personnel in identifying components of a multiple-
plate job and maintaining consistent output during production. See 14.11.3, “Printer’s Marks,” for further
discussion.
"""
from borb.pdf.canvas.layout.annotation.annotation import Annotation


class PrinterMarkAnnotation(Annotation):
    """
    A printer’s mark annotation (PDF 1.4) represents a graphic symbol, such as a registration target, colour bar, or
    cut mark, that may be added to a page to assist production personnel in identifying components of a multiple-
    plate job and maintaining consistent output during production. See 14.11.3, “Printer’s Marks,” for further
    discussion.
    """

    def __init__(self):
        super(PrinterMarkAnnotation, self).__init__()
        # TODO



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
(PDF 1.1) Set the current colour space to use for stroking operations. The
operand name shall be a name object. If the colour space is one that can
be specified by a name and no additional parameters (DeviceGray,
DeviceRGB, DeviceCMYK, and certain cases of Pattern), the name may
be specified directly. Otherwise, it shall be a name defined in the
ColorSpace subdictionary of the current resource dictionary (see 7.8.3,
"Resource Dictionaries"); the associated value shall be an array
describing the colour space (see 8.6.3, "Colour Space Families").

The names DeviceGray, DeviceRGB, DeviceCMYK, and Pattern
always identify the corresponding colour spaces directly; they never refer
to resources in the ColorSpace subdictionary.

The CS operator shall also set the current stroking colour to its initial
value, which depends on the colour space:
In a DeviceGray, DeviceRGB, CalGray, or CalRGB colour space, the
initial colour shall have all components equal to 0.0.
In a DeviceCMYK colour space, the initial colour shall be
[ 0.0 0.0 0.0 1.0 ].

In a Lab or ICCBased colour space, the initial colour shall have all
components equal to 0.0 unless that falls outside the intervals specified
by the space’s Range entry, in which case the nearest valid value shall be
substituted.

In an Indexed colour space, the initial colour value shall be 0.
In a Separation or DeviceN colour space, the initial tint value shall be 1.0
for all colorants.
In a Pattern colour space, the initial colour shall be a pattern object that
causes nothing to be painted.
"""
import logging
import typing
from decimal import Decimal
from typing import List

from borb.io.read.types import AnyPDFType, Name
from borb.pdf.canvas.color.color import CMYKColor, GrayColor, RGBColor, Separation
from borb.pdf.canvas.operator.canvas_operator import CanvasOperator

logger = logging.getLogger(__name__)


class SetColorSpaceStroking(CanvasOperator):
    """
    (PDF 1.1) Set the current colour space to use for stroking operations. The
    operand name shall be a name object. If the colour space is one that can
    be specified by a name and no additional parameters (DeviceGray,
    DeviceRGB, DeviceCMYK, and certain cases of Pattern), the name may
    be specified directly. Otherwise, it shall be a name defined in the
    ColorSpace subdictionary of the current resource dictionary (see 7.8.3,
    "Resource Dictionaries"); the associated value shall be an array
    describing the colour space (see 8.6.3, "Colour Space Families").

    The names DeviceGray, DeviceRGB, DeviceCMYK, and Pattern
    always identify the corresponding colour spaces directly; they never refer
    to resources in the ColorSpace subdictionary.

    The CS operator shall also set the current stroking colour to its initial
    value, which depends on the colour space:
    In a DeviceGray, DeviceRGB, CalGray, or CalRGB colour space, the
    initial colour shall have all components equal to 0.0.
    In a DeviceCMYK colour space, the initial colour shall be
    [ 0.0 0.0 0.0 1.0 ].

    In a Lab or ICCBased colour space, the initial colour shall have all
    components equal to 0.0 unless that falls outside the intervals specified
    by the space’s Range entry, in which case the nearest valid value shall be
    substituted.

    In an Indexed colour space, the initial colour value shall be 0.
    In a Separation or DeviceN colour space, the initial tint value shall be 1.0
    for all colorants. In a Pattern colour space, the initial colour shall be a pattern object that
    causes nothing to be painted.
    """

    def __init__(self):
        super().__init__("CS", 1)

    def invoke(
        self,
        canvas_stream_processor: "CanvasStreamProcessor",  # type: ignore [name-defined]
        operands: typing.List[AnyPDFType] = [],
        event_listeners: typing.List["EventListener"] = [],  # type: ignore [name-defined]
    ) -> None:
        """
        Invoke the CS operator
        """

        assert isinstance(operands[0], Name)
        color_space_name: Name = operands[0]
        color_space: List = []

        if color_space_name not in [
            "DeviceGray",
            "DeviceRGB",
            "DeviceCMYK",
            "CalGray",
            "CalRGB",
            "Lab",
            "ICCBased",
            "Indexed",
            "Pattern",
            "Separation",
        ]:
            color_space_name = canvas_stream_processor.get_resource(
                "ColorSpace", color_space_name
            )

        if not isinstance(color_space_name, Name) and isinstance(
            color_space_name, List
        ):
            assert isinstance(color_space_name[0], Name)
            color_space = color_space_name
            color_space_name = color_space_name[0]

        #
        # Device
        #
        canvas = canvas_stream_processor.get_canvas()
        if color_space_name == "DeviceGray":
            canvas.graphics_state.stroke_color_space = color_space_name
            canvas.graphics_state.stroke_color = GrayColor(Decimal(0))
            return
        if color_space_name == "DeviceRGB":
            canvas.graphics_state.stroke_color_space = color_space_name
            canvas.graphics_state.stroke_color = RGBColor(
                Decimal(0), Decimal(0), Decimal(0)
            )
            return
        if color_space_name == "DeviceCMYK":
            canvas.graphics_state.stroke_color_space = color_space_name
            canvas.graphics_state.stroke_color = CMYKColor(
                Decimal(0), Decimal(0), Decimal(0), Decimal(1)
            )
            return

        #
        # CIE-based
        #

        if color_space_name == "CalGray":
            canvas.graphics_state.stroke_color_space = color_space_name
            canvas.graphics_state.stroke_color = GrayColor(Decimal(0))
            return
        if color_space_name == "CalRGB":
            canvas.graphics_state.stroke_color_space = color_space_name
            canvas.graphics_state.stroke_color = RGBColor(
                Decimal(0), Decimal(0), Decimal(0)
            )
            return
        if color_space_name == "Lab":
            canvas.graphics_state.stroke_color_space = color_space_name
            return
        if color_space_name == "ICCBased":
            canvas.graphics_state.stroke_color_space = color_space_name
            canvas.graphics_state.stroke_color = RGBColor(
                Decimal(0), Decimal(0), Decimal(0)
            )
            return

        #
        # Special
        #
        if color_space_name == "Indexed":
            canvas.graphics_state.stroke_color_space = color_space_name
            return
        if color_space_name == "Pattern":
            canvas.graphics_state.stroke_color_space = operands[0]
            return
        if color_space_name == "Separation":
            canvas.graphics_state.stroke_color_space = color_space
            canvas.graphics_state.stroke_color = Separation(color_space, [Decimal(0)])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    This class provides static convenience methods for generating a stylistic blob
"""
import random
import typing
from decimal import Decimal
from math import cos, radians, sin


class BlobFactory:
    """
    This class provides static convenience methods for generating a stylistic blob
    """

    @staticmethod
    def _sum_points(
        P1: typing.Tuple[Decimal, Decimal], P2: typing.Tuple[Decimal, Decimal]
    ) -> typing.Tuple[Decimal, Decimal]:
        x1, y1 = P1
        x2, y2 = P2
        return x1 + x2, y1 + y2

    @staticmethod
    def _multiply_point(
        multiplier: Decimal, P: typing.Tuple[Decimal, Decimal]
    ) -> typing.Tuple[Decimal, Decimal]:
        x, y = P
        return Decimal(x * multiplier), Decimal(y * multiplier)

    @staticmethod
    def _check_if_object_is_polygon(
        points: typing.List[typing.Tuple[Decimal, Decimal]]
    ):
        if points[0] == points[len(points) - 1]:
            return True
        else:
            return False

    @staticmethod
    def _q_point(P1, P2):
        summand1 = BlobFactory._multiply_point(Decimal(0.75), P1)
        summand2 = BlobFactory._multiply_point(Decimal(0.25), P2)
        Q = BlobFactory._sum_points(summand1, summand2)
        return Q

    @staticmethod
    def _r_point(P1, P2):
        summand1 = BlobFactory._multiply_point(Decimal(0.25), P1)
        summand2 = BlobFactory._multiply_point(Decimal(0.75), P2)
        R = BlobFactory._sum_points(summand1, summand2)
        return R

    @staticmethod
    def smooth_closed_polygon(
        points: typing.List[typing.Tuple[Decimal, Decimal]], number_of_refinements: int
    ):
        """
        This function smooths a polygon by using Chaikin's algorithm.
        In 1974, George Chaikin gave a lecture at the University of Utah in which he specified a novel
        procedure for generating curves from a limited number of points. This algorithm is interesting as it was
        one of the first corner cutting or refinement algorithms specified to generate a curve from a set of control
        points, or control polygon.
        :param points:                  the points of the polygon
        :param number_of_refinements:   the number of iterations to do, each iteration makes the polygon more smooth
        :return:                        a smoothed version of the input polygon
        """
        for _ in range(0, number_of_refinements):
            points_next_iter = []
            for num, pt in enumerate(points):
                P1, P2 = (pt, points[(num + 1) % len(points)])
                Q = BlobFactory._q_point(P1, P2)
                R = BlobFactory._r_point(P1, P2)
                points_next_iter.append(Q)
                points_next_iter.append(R)

            # get everything ready for next iteration
            points = points_next_iter

        # return
        return points

    @staticmethod
    def blob(number_of_edges: int) -> typing.List[typing.Tuple[Decimal, Decimal]]:
        """
        This function returns a smoothed n-sided blob shape
        """

        # generate regular polygon
        points = [
            (Decimal(cos(radians(x))), Decimal(sin(radians(x))))
            for x in range(0, 360, int(360 / number_of_edges))
        ]

        # randomly distort polygon
        random_radius = [Decimal(random.randint(1, 10)) for _ in range(0, len(points))]
        points = [(p[0] * r, p[1] * r) for p, r in zip(points, random_radius)]

        # smoothing
        while len(points) < 1024:
            points = BlobFactory.smooth_closed_polygon(points, 2)
        points.append(points[0])

        # return
        return points



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This class is the base implementation of a transformer pattern.
It reads bytes from a PDF document, and converts it to JSON-like datastructures.
"""
import io
import typing
from typing import Optional

from borb.io.read.types import AnyPDFType, Reference
from borb.io.write.conformance_level import ConformanceLevel


class WriteTransformerState:
    """
    This class represents all the meta-information used in the process of persisting a PDF document.
    This includes:
    - the root object (the Document itself)
    - a cache of indirect objects (by id and hash)
    - references that have been resolved (to avoid endless loops)
    - the default compression level
    - etc
    """

    def __init__(
        self,
        destination: Optional[typing.Union[io.BufferedIOBase, io.RawIOBase]] = None,
        root_object: Optional[AnyPDFType] = None,
        conformance_level: typing.Optional[ConformanceLevel] = None,
    ):
        # fmt: off
        self.destination = destination                                                      # this is the destination to write to (file, byte-buffer, etc)
        self.root_object: Optional[AnyPDFType] = root_object                                # this is the root object (PDF)
        self.indirect_objects_by_id: typing.Dict[int, AnyPDFType] = {}                      # these are the indirect objects (by id)
        self.indirect_objects_by_hash: typing.Dict[int, typing.List[AnyPDFType]] = {}       # these are the indirect objects (by hash)
        self.resolved_references: typing.List[Reference] = []                               # these references have already been written
        self.compression_level: int = 9                                                     # default compression level
        self.conformance_level: typing.Optional[ConformanceLevel] = conformance_level       # default conformance level
        self.apply_font_subsetting: bool = False                                            # whether to apply Font subsetting or not
        # fmt: on


class Transformer:
    """
    This class represents the base transformer for persisting a PDF Document.
    It allows you to add child BaseWriteTransformer implementations to handle specific cases,
    such as persisting Image objects, or Dictionary objects, etc.
    """

    def __init__(self):
        self._handlers: typing.List["Transformer"] = []
        self._parent: typing.Optional["Transformer"] = None

    def add_child_transformer(
        self, handler: "Transformer"  # type: ignore [name-defined]
    ) -> "Transformer":  # type: ignore [name-defined]
        """
        This function adds a BaseWriteTransformer to this WriteBaseTransformer.
        Children of this WriteBaseTransformer will be called in turn, and their
        `can_be_transformed` method should return True if the child WriteBaseTransformer
        can handle the object to be transformed. The first child to return True will have its
        `transform` method called.
        This function returns self.
        """
        self._handlers.append(handler)
        handler._parent = self
        return self

    def get_root_transformer(self) -> "Transformer":  # type: ignore [name-defined]
        """
        This function returns the WriteBaseTransformer at the root of this WriteBaseTransformer hierarchy.
        This allows child WriteBaseTransformer implementations to call the root transformer.
        This is useful for instance when transforming a dictionary, where each key/value would then be transformed
        by delegating the call to the root WriteBaseTransformer.
        """
        p = self
        while p._parent is not None:
            p = p._parent
        return p

    def can_be_transformed(self, any: AnyPDFType):
        """
        This function returns True if this WriteBaseTransformer can transform the input object,
        false otherwise
        """
        return False

    def transform(
        self,
        object_to_transform: AnyPDFType,
        context: Optional[WriteTransformerState] = None,
    ):
        """
        This method writes an object (of type AnyPDFType) to a byte stream (specified in the WriteTransformerState)
        """
        # transform object
        return_value = None
        for h in self._handlers:
            if h.can_be_transformed(object_to_transform):
                return_value = h.transform(
                    object_to_transform,
                    context=context,
                )
                break

        # return
        return return_value

    def _start_object(
        self,
        object_to_transform: AnyPDFType,
        context: Optional[WriteTransformerState],
    ):
        """
        This function starts a new direct object by writing
        its reference number followed by "obj" (e.g. "12 0 obj").
        It also does some bookkeeping to ensure the byte offset is stored in the XREF
        """
        # get offset position
        # fmt: off
        assert context is not None, "A WriteTransformerState must be defined in order to write indirect objects."
        assert context.destination is not None, "A WriteTransformerState must be defined in order to write indirect objects."
        # fmt: on
        byte_offset = context.destination.tell()

        # update offset
        ref = object_to_transform.get_reference()  # type: ignore [union-attr]
        assert ref is not None
        assert isinstance(ref, Reference)
        ref.byte_offset = byte_offset

        # write <object number> <generation number> obj
        assert ref.object_number is not None
        context.destination.write(
            bytes(
                "%d %d obj\n" % (ref.object_number, ref.generation_number or 0),
                "latin1",
            )
        )

    def _end_object(
        self,
        object_to_transform: AnyPDFType,
        context: Optional[WriteTransformerState],
    ):
        """
        This function writes the "endobj" bytes whenever a direct object needs to be closed
        """
        # write endobj
        # fmt: off
        assert context is not None, "A WriteTransformerState must be defined in order to write indirect objects."
        assert context.destination is not None, "A WriteTransformerState must be defined in order to write indirect objects."
        # fmt: on
        context.destination.write(bytes("endobj\n\n", "latin1"))

    @staticmethod
    def _hash(obj: typing.Any) -> int:
        h: Optional[int] = None
        # hash
        try:
            h = hash(obj)
        except:
            pass
        # __hash__
        try:
            h = obj.__hash__()
        except:
            pass
        if h is None:
            raise TypeError("unhashable type: %s" % obj.__class__.__name__)
        return h

    def get_reference(
        self, object: AnyPDFType, context: WriteTransformerState
    ) -> Reference:
        """
        This function builds a Reference for the input object
        References are re-used whenever possible (hashing is used to detect duplicate objects)
        """
        is_unique: bool = False
        try:
            is_unique = object.is_unique()  # type: ignore[union-attr]
        except:
            pass

        # look up object by ID
        obj_id = id(object)
        if obj_id in context.indirect_objects_by_id:
            cached_indirect_object: AnyPDFType = context.indirect_objects_by_id[obj_id]
            assert not isinstance(cached_indirect_object, Reference)
            return cached_indirect_object.get_reference()  # type: ignore[union-attr]

        # look through existing indirect object hashes
        obj_hash: int = self._hash(object)
        if (not is_unique) and obj_hash in context.indirect_objects_by_hash:
            for obj in context.indirect_objects_by_hash[obj_hash]:
                if obj == object:
                    ref = obj.get_reference()  # type: ignore [union-attr]
                    assert ref is not None
                    assert isinstance(ref, Reference)
                    object.set_reference(ref)  # type: ignore [union-attr]
                    return ref

        # generate new object number
        existing_obj_numbers = set(
            [
                item.get_reference().object_number  # type: ignore [union-attr]
                for sublist in [v for k, v in context.indirect_objects_by_hash.items()]
                for item in sublist
            ]
        )
        obj_number = len(existing_obj_numbers) + 1
        while obj_number in existing_obj_numbers:  # type: ignore [union-attr]
            obj_number += 1

        # build reference
        ref = Reference(object_number=obj_number)
        object.set_reference(ref)  # type: ignore [union-attr]

        # insert into context.indirect_objects_by_hash
        if obj_hash in context.indirect_objects_by_hash:
            context.indirect_objects_by_hash[obj_hash].append(object)
        else:
            context.indirect_objects_by_hash[obj_hash] = [object]

        # insert into context.indirect_objects_by_id
        context.indirect_objects_by_id[obj_id] = object

        # return
        return ref



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This implementation of LayoutElement aggregates other LayoutElements
and lays them out on consecutive lines. If a line is full, it overflows
into the next line.
"""
import typing
from decimal import Decimal

from borb.pdf.canvas.geometry.rectangle import Rectangle
from borb.pdf.canvas.layout.layout_element import LayoutElement


class InlineFlow(LayoutElement):
    """
    This implementation of LayoutElement aggregates other LayoutElements
    and lays them out on consecutive lines. If a line is full, it overflows
    into the next line.
    """

    def __init__(self):
        super(InlineFlow, self).__init__()
        self._content: typing.List[LayoutElement] = []

    def add(self, e: LayoutElement) -> "InlineFlow":
        """
        This function adds a LayoutElement to this InlineFlow
        :param e:   the LayoutElement to be added
        :return:    self
        """
        if isinstance(e, InlineFlow):
            for child_e in e._content:
                self.add(child_e)
        else:
            self._content.append(e)
        return self

    def extend(self, es: typing.List[LayoutElement]) -> "InlineFlow":
        """
        This function adds a typing.List of LayoutElement(s) to this InlineFlow
        :param es:   the LayoutElements to be added
        :return:    self
        """
        for e in es:
            self.add(e)
        return self

    @staticmethod
    def _get_min_content_box(e: LayoutElement) -> Rectangle:
        from borb.pdf import TableCell

        c: TableCell = TableCell(e)
        c._calculate_min_and_max_layout_box()
        return Rectangle(Decimal(0), Decimal(0), c._min_width, c._max_height)

    def _get_content_box(self, available_space: Rectangle) -> Rectangle:

        # all lines
        layout_lines: typing.List[typing.List[LayoutElement]] = [[]]
        layout_lines_height: typing.List[Decimal] = [Decimal(0)]

        # current line
        layout_line_width: Decimal = Decimal(0)

        # distribute content over lines
        for e in self._content:
            cbox: Rectangle = InlineFlow._get_min_content_box(e)
            w: Decimal = cbox.get_width()
            h: Decimal = cbox.get_height()
            if layout_line_width + w > available_space.get_width() or (
                e.__class__.__name__ == "LineBreakChunk"
            ):
                e._previous_layout_box = Rectangle(
                    available_space.get_x(), Decimal(0), w, h
                )
                layout_lines.append([e])
                layout_lines_height.append(h)
                layout_line_width = w
            else:
                e._previous_layout_box = Rectangle(
                    available_space.get_x() + layout_line_width, Decimal(0), w, h
                )
                layout_lines[-1].append(e)
                layout_line_width += w
                layout_lines_height[-1] = max(layout_lines_height[-1], h)

        # set height on all elements
        y: Decimal = available_space.get_y() + available_space.get_height()
        for i, line in enumerate(layout_lines):
            y -= layout_lines_height[i]
            for e in line:
                e._previous_layout_box = Rectangle(
                    e._previous_layout_box.get_x(),
                    y,
                    e._previous_layout_box.get_width(),
                    layout_lines_height[i],
                )

        # calculate total dimensions
        return Rectangle(
            available_space.get_x(),
            available_space.get_y()
            + available_space.get_height()
            - sum(layout_lines_height),
            available_space.get_width(),
            sum(layout_lines_height),
        )

    def _paint_content_box(self, page: "Page", content_box: Rectangle) -> None:
        for e in self._content:
            e.paint(page, e.get_previous_layout_box())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
(PDF 1.2) Same as SC but also supports Pattern, Separation, DeviceN
and ICCBased colour spaces.
"""
import typing
from decimal import Decimal
from typing import List

from borb.io.read.types import AnyPDFType
from borb.pdf.canvas.color.color import CMYKColor, GrayColor, RGBColor, Separation
from borb.pdf.canvas.operator.canvas_operator import CanvasOperator


class SetColorStroking(CanvasOperator):
    """
    (PDF 1.2) Same as SC but also supports Pattern, Separation, DeviceN
    and ICCBased colour spaces.
    If the current stroking colour space is a Separation, DeviceN, or
    ICCBased colour space, the operands c 1 ... c n shall be numbers. The
    number of operands and their interpretation depends on the colour space.
    If the current stroking colour space is a Pattern colour space, name shall
    be the name of an entry in the Pattern subdictionary of the current
    resource dictionary (see 7.8.3, "Resource Dictionaries"). For an
    uncoloured tiling pattern (PatternType = 1 and PaintType = 2), c 1 ... c n
    shall be component values specifying a colour in the pattern’s underlying
    colour space. For other types of patterns, these operands shall not be
    specified.
    """

    def __init__(self, canvas_stream_processor: "CanvasStreamProcessor"):  # type: ignore [name-defined]
        super().__init__("SCN", 0)
        self._canvas = canvas_stream_processor.get_canvas()

    def get_number_of_operands(self) -> int:
        """
        This function returns the number of operands for the SCN operator.
        The number of operands and their interpretation depends on the colour space.
        """
        stroke_color_space = self._canvas.graphics_state.stroke_color_space
        if stroke_color_space == "DeviceCMYK":
            return 4
        if stroke_color_space == "DeviceGray":
            return 1
        if stroke_color_space == "DeviceRGB":
            return 3
        # separation
        if (
            isinstance(stroke_color_space, List)
            and len(stroke_color_space) == 4
            and stroke_color_space[0] == "Separation"
        ):
            return 1
        return self._number_of_operands

    def invoke(
        self,
        canvas_stream_processor: "CanvasStreamProcessor",  # type: ignore [name-defined]
        operands: typing.List[AnyPDFType] = [],
        event_listeners: typing.List["EventListener"] = [],  # type: ignore [name-defined]
    ) -> None:
        """
        Invoke the SCN operator
        """
        canvas = canvas_stream_processor.get_canvas()
        stroke_color_space = canvas.graphics_state.stroke_color_space
        if stroke_color_space == "DeviceCMYK":
            assert isinstance(
                operands[0], Decimal
            ), "Operand 0 of SCN must be a Decimal"
            assert isinstance(
                operands[1], Decimal
            ), "Operand 1 of SCN must be a Decimal"
            assert isinstance(
                operands[2], Decimal
            ), "Operand 2 of SCN must be a Decimal"
            assert isinstance(
                operands[3], Decimal
            ), "Operand 3 of SCN must be a Decimal"
            canvas.graphics_state.stroke_color = CMYKColor(
                operands[0],
                operands[1],
                operands[2],
                operands[3],
            )
            return

        if stroke_color_space == "DeviceGray":
            assert isinstance(
                operands[0], Decimal
            ), "Operand 0 of SCN must be a Decimal"
            canvas.graphics_state.stroke_color = GrayColor(operands[0])
            return

        if stroke_color_space == "DeviceRGB":
            assert isinstance(
                operands[0], Decimal
            ), "Operand 0 of SCN must be a Decimal"
            assert isinstance(
                operands[1], Decimal
            ), "Operand 1 of SCN must be a Decimal"
            assert isinstance(
                operands[2], Decimal
            ), "Operand 2 of SCN must be a Decimal"
            canvas.graphics_state.stroke_color = RGBColor(
                operands[0],
                operands[1],
                operands[2],
            )
            return

        # separation
        if (
            isinstance(stroke_color_space, List)
            and stroke_color_space[0] == "Separation"
        ):
            assert isinstance(
                operands[0], Decimal
            ), "Operand 0 of SCN must be a Decimal"
            canvas.graphics_state.stroke_color = Separation(
                canvas.graphics_state.stroke_color_space, [operands[0]]
            )
            return



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
The cross-reference table contains information that permits random access to indirect objects within the file so
that the entire file need not be read to locate any particular object.
This implementation of PlainTextXREF is invoked whenever the XREF could not be loaded properly.
This implementation will then loop over the entire PDF file to find obj declaration statements.
"""
import io
import logging
import typing

from borb.io.read.tokenize.high_level_tokenizer import HighLevelTokenizer
from borb.io.read.types import Reference, Name
from borb.pdf.xref.plaintext_xref import PlainTextXREF
from borb.pdf.xref.xref import XREF

logger = logging.getLogger(__name__)


class RebuiltXREF(PlainTextXREF):
    """
    The cross-reference table contains information that permits random access to indirect objects within the file so
    that the entire file need not be read to locate any particular object.
    This implementation of PlainTextXREF is invoked whenever the XREF could not be loaded properly.
    This implementation will then loop over the entire PDF file to find obj declaration statements.
    """

    def read(
        self,
        src: typing.Union[io.BufferedIOBase, io.RawIOBase, io.BytesIO],
        tok: HighLevelTokenizer,
        initial_offset: typing.Optional[int] = None,
    ) -> "XREF":
        """
        This method attempts to read a plaintext XREF from the given io_source.
        It will either throw an exception, or return this XREF
        """

        # logger
        logger.info("Switching to RebuiltXREF parsing.")

        # get pos
        pos_before: int = src.tell()

        # set pos
        src.seek(0)

        # get object declarations
        i: int = 0
        trailer_pos: typing.Optional[int] = None
        bytes_in_pdf: typing.Optional[bytes] = src.read()
        assert (
            bytes_in_pdf is not None
        ), "rebuilding an XREF is only possible if all the bytes of the PDF are known"
        while i < len(bytes_in_pdf):
            # 0 0 obj
            if (
                48 <= bytes_in_pdf[i] <= 57  # <number>
                and bytes_in_pdf[i + 1] == 32  # <space>
                and 48 <= bytes_in_pdf[i + 2] <= 57  # <number>
                and bytes_in_pdf[i + 3] == 32  # <space>
                and bytes_in_pdf[i + 4] == 111  # 'o'
                and bytes_in_pdf[i + 5] == 98  # 'b'
                and bytes_in_pdf[i + 6] == 106  # 'j'
            ):
                logger.debug(
                    "%d %d obj at %d"
                    % (bytes_in_pdf[i] - 48, bytes_in_pdf[i + 1] - 48, i)
                )
                self._entries.append(
                    Reference(
                        object_number=bytes_in_pdf[i] - 48,
                        generation_number=bytes_in_pdf[i + 2] - 48,
                        byte_offset=i,
                    )
                )
                i += 7
                continue

            # 00 0 obj
            if (
                48 <= bytes_in_pdf[i] <= 57  # <number>
                and 48 <= bytes_in_pdf[i + 1] <= 57  # <number>
                and bytes_in_pdf[i + 2] == 32  # <space>
                and 48 <= bytes_in_pdf[i + 3] <= 57  # <number>
                and bytes_in_pdf[i + 4] == 32  # <space>
                and bytes_in_pdf[i + 5] == 111  # 'o'
                and bytes_in_pdf[i + 6] == 98  # 'b'
                and bytes_in_pdf[i + 7] == 106  # 'j'
            ):
                logger.debug(
                    "%d %d obj at %d"
                    % (
                        (bytes_in_pdf[i] - 48) * 10 + (bytes_in_pdf[i + 1] - 48),
                        bytes_in_pdf[i + 3] - 48,
                        i,
                    )
                )
                self._entries.append(
                    Reference(
                        object_number=(bytes_in_pdf[i] - 48) * 10
                        + (bytes_in_pdf[i + 1] - 48),
                        generation_number=bytes_in_pdf[i + 3] - 48,
                        byte_offset=i,
                    )
                )
                i += 8
                continue

            # 000 0 obj
            if (
                48 <= bytes_in_pdf[i] <= 57  # <number>
                and 48 <= bytes_in_pdf[i + 1] <= 57  # <number>
                and 48 <= bytes_in_pdf[i + 2] <= 57  # <number>
                and bytes_in_pdf[i + 3] == 32  # <space>
                and 48 <= bytes_in_pdf[i + 4] <= 57  # <number>
                and bytes_in_pdf[i + 5] == 32  # <space>
                and bytes_in_pdf[i + 6] == 111  # 'o'
                and bytes_in_pdf[i + 7] == 98  # 'b'
                and bytes_in_pdf[i + 8] == 106  # 'j'
            ):
                obj_nr: int = (
                    (bytes_in_pdf[i] - 48) * 100
                    + (bytes_in_pdf[i + 1] - 48) * 10
                    + (bytes_in_pdf[i + 2] - 48)
                )
                logger.debug("%d %d obj at %d" % (obj_nr, bytes_in_pdf[i + 4] - 48, i))
                self._entries.append(
                    Reference(
                        object_number=obj_nr,
                        generation_number=bytes_in_pdf[i + 4] - 48,
                        byte_offset=i,
                    )
                )
                i += 9
                continue

            # trailer
            if (
                bytes_in_pdf[i] == 116  # 't'
                and bytes_in_pdf[i + 1] == 114  # 'r'
                and bytes_in_pdf[i + 2] == 97  # 'a'
                and bytes_in_pdf[i + 3] == 105  # 'i'
                and bytes_in_pdf[i + 4] == 108  # 'l'
                and bytes_in_pdf[i + 5] == 101  # 'e'
                and bytes_in_pdf[i + 6] == 114  # 'r'
            ):
                trailer_pos = i

            # default
            i += 1

        # read trailer
        assert trailer_pos is not None
        src.seek(trailer_pos)
        self[Name("Trailer")] = self._read_trailer(src, tok)
        self[Name("Trailer")].set_parent(self)

        # restore pos
        src.seek(pos_before)

        # return
        return self



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
An ink annotation (PDF 1.3) represents a freehand “scribble” composed of one or more disjoint paths. When
opened, it shall display a pop-up window containing the text of the associated note. Table 182 shows the
annotation dictionary entries specific to this type of annotation.
"""
from borb.pdf.canvas.layout.annotation.annotation import Annotation


class InkAnnotation(Annotation):
    """
    An ink annotation (PDF 1.3) represents a freehand “scribble” composed of one or more disjoint paths. When
    opened, it shall display a pop-up window containing the text of the associated note. Table 182 shows the
    annotation dictionary entries specific to this type of annotation.
    """

    def __init__(self):
        super(InkAnnotation, self).__init__()
        # TODO



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    CID-keyed fonts provide a convenient and efficient method for defining multiple-byte character encodings and
   fonts with a large number of glyphs. These capabilities provide great flexibility for representing text in writing
    systems for languages with large character sets, such as Chinese, Japanese, and Korean (CJK).

    The CID-keyed font architecture specifies the external representation of certain font programs, called CMap
    and CIDFont files, along with some conventions for combining and using those files. As mentioned earlier, PDF
    does not support the entire CID-keyed font architecture, which is independent of PDF; CID-keyed fonts may be
    used in other environments.

    A Type 0 CIDFont contains glyph descriptions based on CFF
"""
import typing

from borb.io.read.types import Decimal as bDecimal
from borb.io.read.types import List, Name
from borb.pdf.canvas.font.font import Font


class CIDType0Font(Font):
    """
    A Type 0 CIDFont contains glyph descriptions based on CFF
    """

    def __init__(self):
        super(CIDType0Font, self).__init__()
        self._width_cache: typing.Dict[int, bDecimal] = {}

    def get_width(self, character_identifier: int) -> typing.Optional[bDecimal]:
        """
        This function returns the width (in text space) of a given character identifier.
        If this Font is unable to represent the glyph that corresponds to the character identifier,
        this function returns None
        """

        # check cache
        if character_identifier in self._width_cache:
            return self._width_cache[character_identifier]

        # Default value: none (the DW value shall be used for all glyphs).
        dw: bDecimal = self["DW"] if "DW" in self else bDecimal(1000)
        if "W" not in self:
            return dw

        assert "W" in self
        assert isinstance(self["W"], List)
        i: int = 0
        cid: int = 0
        cid_width: int = 0
        while i < len(self["W"]):
            # <char_start_code> [<width>+]
            if (
                isinstance(self["W"][i], bDecimal)
                and i + 1 < len(self["W"])
                and isinstance(self["W"][i + 1], List)
            ):
                for j in range(0, len(self["W"][i + 1])):
                    cid = int(self["W"][i]) + j
                    cid_width = int(self["W"][i + 1][j])
                    self._width_cache[cid] = bDecimal(cid_width)
                i += 2
                continue
            # <char_start_code> <char_end_code> <width>
            if (
                isinstance(self["W"][i], bDecimal)
                and i + 2 < len(self["W"])
                and isinstance(self["W"][i + 1], bDecimal)
                and isinstance(self["W"][i + 2], bDecimal)
            ):
                for j in range(int(self["W"][i]), int(self["W"][i + 1]) + 1):
                    cid = j
                    cid_width = int(self["W"][i + 2])
                    self._width_cache[cid] = bDecimal(cid_width)
                i += 3
                continue

        # check cache
        if character_identifier in self._width_cache:
            return self._width_cache[character_identifier]

        # default
        return dw

    def get_ascent(self) -> bDecimal:
        """
        This function returns the maximum height above the baseline reached by glyphs in this font.
        The height of glyphs for accented characters shall be excluded.
        """
        assert "FontDescriptor" in self
        assert "Ascent" in self["FontDescriptor"]
        return self["FontDescriptor"]["Ascent"]

    def get_descent(self) -> bDecimal:
        """
        This function returns the maximum depth below the baseline reached by glyphs in this font.
        The value shall be a negative number.
        """
        assert "FontDescriptor" in self
        assert "Descent" in self["FontDescriptor"]
        return self["FontDescriptor"]["Descent"]

    def _empty_copy(self) -> "Font":
        return CIDType0Font()

    def __deepcopy__(self, memodict={}):
        # fmt: off
        f_out: CIDType0Font = super(CIDType0Font, self).__deepcopy__(memodict)
        f_out[Name("Subtype")] = Name("CIDFontType0")
        f_out._width_cache: typing.Dict[int, bDecimal] = {k: v for k, v in self._width_cache.items()}
        return f_out
        # fmt: on



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    This implementation of Event is triggered when an Image has been processed using a Do instruction
"""
from decimal import Decimal

from PIL import Image as PILImage  # type: ignore [import]

from borb.pdf.canvas.canvas_graphics_state import CanvasGraphicsState
from borb.pdf.canvas.event.event_listener import Event


class ImageRenderEvent(Event):
    """
    This implementation of Event is triggered when an Image has been processed using a Do instruction
    """

    def __init__(self, graphics_state: CanvasGraphicsState, image: PILImage):  # type: ignore[valid-type]
        self._image: PILImage = image  # type: ignore[valid-type]

        # calculate position
        v = graphics_state.ctm.cross(Decimal(0), Decimal(0), Decimal(1))
        self._x: Decimal = v[0]
        self._y: Decimal = v[1]

        # calculate display size
        v = graphics_state.ctm.cross(Decimal(1), Decimal(1), Decimal(0))
        self._width: Decimal = max(abs(v[0]), Decimal(1))
        self._height: Decimal = max(abs(v[1]), Decimal(1))

    def get_image(self) -> PILImage:  # type: ignore[valid-type]
        """
        Get the (source) Image
        This Image may have different dimensions than
        how it is displayed in the PDF
        """
        return self._image

    def get_x(self) -> Decimal:
        """
        Get the x-coordinate at which the Image is drawn
        """
        return self._x

    def get_y(self) -> Decimal:
        """
        Get the y-coordinate at which the Image is drawn
        """
        return self._y

    def get_width(self) -> Decimal:
        """
        Get the width of the (scaled) Image
        """
        return self._width

    def get_height(self) -> Decimal:
        """
        Get the height of the (scaled) Image
        """
        return self._height



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Move to the next line and show a text string. This operator shall have the
same effect as the code
T*
string Tj
"""
import typing

from borb.io.read.types import AnyPDFType
from borb.pdf.canvas.operator.canvas_operator import CanvasOperator


class MoveToNextLineShowText(CanvasOperator):
    """
    Move to the next line and show a text string. This operator shall have the
    same effect as the code
    T*
    string Tj
    """

    def __init__(self):
        super().__init__("'", 1)

    def invoke(
        self,
        canvas_stream_processor: "CanvasStreamProcessor",  # type: ignore [name-defined]
        operands: typing.List[AnyPDFType] = [],
        event_listeners: typing.List["EventListener"] = [],  # type: ignore [name-defined]
    ) -> None:
        """
        Invoke the ' operator
        """
        move_to_next_line_op: typing.Optional[
            CanvasOperator
        ] = canvas_stream_processor.get_operator("T*")
        assert (
            move_to_next_line_op
        ), "Operator T* must be defined for operator ' to function."
        move_to_next_line_op.invoke(canvas_stream_processor, [], event_listeners)

        show_text_op: typing.Optional[
            CanvasOperator
        ] = canvas_stream_processor.get_operator("Tj")
        assert show_text_op, "Operator Tj must be defined for operator ' to function"
        show_text_op.invoke(canvas_stream_processor, operands, event_listeners)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This class represents a trie[str, typing.Any]
"""
import typing


class Trie:
    """
    This class represents a trie[str, typing.Any]
    """

    class TrieNode:
        """
        This class represents a node in a trie
        """

        def __init__(self, value: typing.Optional[typing.Any] = None):
            self._children: typing.Dict[str, "Trie.TrieNode"] = {}
            self._value: typing.Optional[typing.Any] = value

        def __len__(self) -> int:
            return (0 if self._value is None else 1) + sum(
                [len(v) for k, v in self._children.items()]
            )

    def __init__(self):
        self._root: typing.Optional[Trie.TrieNode] = None

    def __len__(self) -> int:
        return 0 if self._root is None else len(self._root)

    def __getitem__(self, item) -> typing.Optional[typing.Any]:
        n: typing.Optional[Trie.TrieNode] = self._root
        if n is None:
            return None
        for c in item:
            if c in n._children:
                n = n._children[c]
            else:
                return None
        assert n is not None, "unexpected error while performing __getitem__ on Trie"
        return n._value

    def __setitem__(self, key, value):
        n: typing.Optional[Trie.TrieNode] = self._root
        if n is None:
            self._root = Trie.TrieNode()
            n = self._root
        assert n is not None, "unexpected error while performing __setitem__ on Trie"
        for c in key:
            if c not in n._children:
                n._children[c] = Trie.TrieNode()
            n = n._children[c]
        assert n is not None, "unexpected error while performing __setitem__ on Trie"
        n._value = value
        return self



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This module contains everything needed to implement a LayoutElement representing a barcode.
"""
import typing
from decimal import Decimal
from enum import Enum

import barcode  # type: ignore [import]
import qrcode  # type: ignore [import]
from PIL import Image as PILImage  # type: ignore [import]
from barcode.writer import ImageWriter as BarcodeImageWriter  # type: ignore [import]

from borb.pdf.canvas.color.color import Color, HexColor
from borb.pdf.canvas.layout.image.image import Image
from borb.pdf.canvas.layout.layout_element import Alignment


class BarcodeType(Enum):
    """
    This Enum represents the various types of supported barcodes
    """

    CODE_128 = "code128"
    CODE_39 = "code39"
    EAN = "ean"
    EAN_13 = "ean13"
    EAN_14 = "ean14"
    EAN_8 = "ean8"
    GS_1 = "isbn13"
    GS_128 = "gs1_128"
    GTIN = "ean14"
    ISBN = "isbn13"
    ISBN_10 = "isbn10"
    ISBN_13 = "isbn13"
    ISSN = "issn"
    ITF = "itf"
    JAN = "jan"
    PZN = "pzn"
    QR = "qr"
    UPC = "upca"
    UPC_A = "upca"


class InMemoryBarcodeWriter(BarcodeImageWriter):
    """
    This class inherits from BarcodeImageWriter, and enables
    access to the PILImage being built
    """

    def __init__(self):
        super(InMemoryBarcodeWriter, self).__init__(format="JPEG", mode="RGB")
        self.output_image: typing.Optional[PILImage] = None

    def save(self, filename, output):
        """
        Saves the rendered output to `filename` storing the output.

        :parameters:
            filename : String
                Filename without extension.
            output : String
                The rendered output.

        :returns: The full filename with extension.
        :rtype: String
        """
        self.output_image = output

    def get_output_image(self) -> PILImage:  # type: ignore[valid-type]
        """
        This function returns the PILImage representing the barcode
        """
        return self.output_image


class Barcode(Image):
    """
    This implementation of LayoutElement represents a barcode.
    """

    def __init__(
        self,
        data: str,
        type: BarcodeType,
        stroke_color: Color = HexColor("000000"),
        fill_color: Color = HexColor("ffffff"),
        border_bottom: bool = False,
        border_color: Color = HexColor("000000"),
        border_left: bool = False,
        border_radius_bottom_left: Decimal = Decimal(0),
        border_radius_bottom_right: Decimal = Decimal(0),
        border_radius_top_left: Decimal = Decimal(0),
        border_radius_top_right: Decimal = Decimal(0),
        border_right: bool = False,
        border_top: bool = False,
        border_width: Decimal = Decimal(1),
        height: typing.Optional[Decimal] = None,
        horizontal_alignment: Alignment = Alignment.LEFT,
        margin_bottom: Decimal = Decimal(0),
        margin_left: Decimal = Decimal(0),
        margin_right: Decimal = Decimal(0),
        margin_top: Decimal = Decimal(0),
        padding_bottom: Decimal = Decimal(0),
        padding_left: Decimal = Decimal(0),
        padding_right: Decimal = Decimal(0),
        padding_top: Decimal = Decimal(0),
        vertical_alignment: Alignment = Alignment.TOP,
        width: typing.Optional[Decimal] = None,
    ):
        self._data: str = data
        self._type: BarcodeType = type
        self._stroke_color: Color = stroke_color
        self._fill_color: Color = fill_color

        assert stroke_color != fill_color

        if type == BarcodeType.QR:
            image = self._generate_qr_code(data)
        else:
            image = self._generate_image_except_qr_code(data, type)

        # call to super
        super(Barcode, self).__init__(
            image,
            width=width,
            height=height,
            border_bottom=border_bottom,
            border_color=border_color,
            border_left=border_left,
            border_radius_bottom_left=border_radius_bottom_left,
            border_radius_bottom_right=border_radius_bottom_right,
            border_radius_top_left=border_radius_top_left,
            border_radius_top_right=border_radius_top_right,
            border_right=border_right,
            border_top=border_top,
            border_width=border_width,
            horizontal_alignment=horizontal_alignment,
            margin_bottom=margin_bottom,
            margin_left=margin_left,
            margin_right=margin_right,
            margin_top=margin_top,
            padding_bottom=padding_bottom,
            padding_left=padding_left,
            padding_right=padding_right,
            padding_top=padding_top,
            vertical_alignment=vertical_alignment,
        )
        self._background_color = fill_color

    def _generate_image_except_qr_code(self, data: str, type: BarcodeType):
        # generate image using barcode library
        writer: InMemoryBarcodeWriter = InMemoryBarcodeWriter()
        ean = barcode.get(name=type.value, code=data, writer=writer)
        ean.save(
            "",
            options={
                "foreground": self._stroke_color.to_rgb().to_hex_string(),
                "background": self._fill_color.to_rgb().to_hex_string(),
            },
        )

        # get the rendered image from InMemoryBarcodeWriter
        image: PILImage = writer.get_output_image()  # type: ignore[valid-type]
        assert image is not None
        assert image.width > 0
        assert image.height > 0

        # return
        return image

    def _generate_qr_code(self, data: str):
        qr = qrcode.QRCode(
            version=None,
            error_correction=qrcode.constants.ERROR_CORRECT_L,
            box_size=10,
            border=4,
        )
        qr.add_data(data)
        qr.make(fit=True)

        # png to jpg
        png_image: PILImage = qr.make_image(  # type: ignore[valid-type]
            fill_color=self._stroke_color.to_rgb().to_hex_string(),
            back_color=self._fill_color.to_rgb().to_hex_string(),
        )
        jpg_image = png_image.convert("RGB")  # type: ignore [attr-defined]

        return jpg_image



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
A movie annotation (PDF 1.2) contains animated graphics and sound to be presented on the computer screen
and through the speakers. When the annotation is activated, the movie shall be played. Table 186 shows the
annotation dictionary entries specific to this type of annotation. Movies are discussed in 13.4, “Movies.”
"""
from borb.pdf.canvas.layout.annotation.annotation import Annotation


class MovieAnnotation(Annotation):
    """
    A movie annotation (PDF 1.2) contains animated graphics and sound to be presented on the computer screen
    and through the speakers. When the annotation is activated, the movie shall be played. Table 186 shows the
    annotation dictionary entries specific to this type of annotation. Movies are discussed in 13.4, “Movies.”
    """

    def __init__(self):
        super(MovieAnnotation, self).__init__()
        # TODO



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Decompresses data encoded using the LZW (Lempel-Ziv-Welch)
adaptive compression method, reproducing the original
text or binary data.
"""
import typing


class bitarray:
    """
    This class works allows you to work with a bytes input
    and retrieve individual bits (as integers)
    """

    def __init__(self, input: bytes):
        self._src: bytes = input
        self._pos: int = -1
        self._buffer: typing.List[int] = []
        self._default_to_return: int = 256

    def _read_next_byte(self):
        self._pos += 1
        self._buffer += [int(x) for x in "{0:08b}".format(self._src[self._pos])]

    def next(self, n) -> int:
        """
        This function reads the next n bits from the input
        :param n:   the number of bits to retrieve
        :return:    the next n bits from the input as an integer
        """
        try:
            while n > len(self._buffer):
                self._read_next_byte()
            x: typing.List[int] = self._buffer[:n]
            self._buffer = self._buffer[n:]
            return int("".join([str(y) for y in x]), 2)
        except:
            return self._default_to_return


class LZWDecode:
    """
    Decompresses data encoded using the LZW (Lempel-Ziv-
    Welch) adaptive compression method, reproducing the original
    text or binary data.
    """

    def __init__(self):
        self._lookup_table: typing.Dict[int, bytearray] = {}
        self._table_index: int = 0
        self._bits_to_read: int = 9

    def _init_lookup_table(self):
        self._lookup_table = {i: i.to_bytes(1, "big") for i in range(0, 256)}
        self._table_index = 258
        self._bits_to_read = 9

    def _add_to_lookup_table(self, prev_bytes: bytearray, new_bytes: bytes):
        self._lookup_table[self._table_index] = prev_bytes + new_bytes
        self._table_index += 1
        if self._table_index == 511:
            self._bits_to_read = 10
        elif self._table_index == 1023:
            self._bits_to_read = 11
        elif self._table_index == 2047:
            self._bits_to_read = 12

    def decode(self, input: bytes):
        """
        Decompresses data encoded using the LZW (Lempel-Ziv-Welch)
        adaptive compression method
        """

        # output
        bytes_out: bytearray = bytearray()

        # read
        bit_input: bitarray = bitarray(input)
        prev_code: int = 0
        code: int = 0

        while code != 257:
            code = bit_input.next(self._bits_to_read)
            if code == 257:
                break

            # init
            if code == 256:
                self._init_lookup_table()
                code = bit_input.next(self._bits_to_read)
                if code == 257:
                    break
                bytes_out += self._lookup_table[code]
                prev_code = code
                continue

            # normal behaviour
            x: bytearray = bytearray()
            if code < self._table_index:
                x = self._lookup_table[code]
                bytes_out += x
                self._add_to_lookup_table(self._lookup_table[prev_code], x[0:1])
                prev_code = code
            else:
                x = self._lookup_table[prev_code]
                x = x + x[0:1]
                bytes_out += x
                self._add_to_lookup_table(x, bytearray())
                prev_code = code

        # return bytes
        return bytes_out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This implementation of LayoutElement represents an horizontal line across the page
"""
import typing
from decimal import Decimal

from borb.pdf.canvas.color.color import Color, HexColor, RGBColor
from borb.pdf.canvas.geometry.rectangle import Rectangle
from borb.pdf.canvas.layout.layout_element import LayoutElement
from borb.pdf.page.page import Page


class HorizontalRule(LayoutElement):
    """
    This implementation of LayoutElement represents an horizontal line across the page
    """

    def __init__(
        self,
        line_width: Decimal = Decimal(1),
        line_color: Color = HexColor("000000"),
        margin_top: typing.Optional[Decimal] = None,
        margin_bottom: typing.Optional[Decimal] = None,
    ):
        # fmt: off
        super(HorizontalRule, self).__init__(
            margin_top=margin_top if margin_top is not None else Decimal(5),
            margin_bottom=margin_bottom if margin_bottom is not None else Decimal(5),
        )
        # fmt: on
        self._line_width: Decimal = line_width
        self._line_color: Color = line_color

    def _get_content_box(self, available_space: Rectangle) -> Rectangle:
        return Rectangle(
            available_space.get_x(),
            available_space.get_y() + available_space.get_height() - self._line_width,
            available_space.get_width(),
            self._line_width,
        )

    def _paint_content_box(self, page: Page, available_space: Rectangle) -> None:

        # write l operator
        rgb_color: RGBColor = self._line_color.to_rgb()
        content = " q %f %f %f RG %f %f m %f %f l s Q " % (
            float(rgb_color.red),
            float(rgb_color.green),
            float(rgb_color.blue),
            float(available_space.get_x()),
            float(
                available_space.get_y()
                + available_space.get_height()
                - self._line_width
            ),
            float(available_space.get_x() + available_space.get_width()),
            float(
                available_space.get_y()
                + available_space.get_height()
                - self._line_width
            ),
        )

        # modify content stream
        page._append_to_content_stream(content)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Text markup annotations shall appear as highlights, underlines, strikeouts (all PDF 1.3), or jagged (“squiggly”)
underlines (PDF 1.4) in the text of a document. When opened, they shall display a pop-up window containing
the text of the associated note. Table 179 shows the annotation dictionary entries specific to these types of
annotations.
"""
import zlib
from decimal import Decimal

from borb.io.read.types import Name, Stream, Dictionary, List, Decimal as bDecimal
from borb.pdf.canvas.color.color import HexColor, Color
from borb.pdf.canvas.geometry.rectangle import Rectangle
from borb.pdf.canvas.layout.annotation.annotation import Annotation


class SquigglyAnnotation(Annotation):
    """
    Text markup annotations shall appear as highlights, underlines, strikeouts (all PDF 1.3), or jagged (“squiggly”)
    underlines (PDF 1.4) in the text of a document. When opened, they shall display a pop-up window containing
    the text of the associated note. Table 179 shows the annotation dictionary entries specific to these types of
    annotations.
    """

    def __init__(
        self,
        bounding_box: Rectangle,
        stroke_width: Decimal = Decimal(1),
        stroke_color: Color = HexColor("ff0000"),
    ):
        super(SquigglyAnnotation, self).__init__(bounding_box)

        # (Required) The type of annotation that this dictionary describes; shall
        # be Redact for a redaction annotation.
        self[Name("Subtype")] = Name("Squiggly")

        # (Optional; PDF 1.2) An appearance dictionary specifying how the
        # annotation shall be presented visually on the page (see 12.5.5,
        # “Appearance Streams”). Individual annotation handlers may ignore this
        # entry and provide their own appearances.
        self[Name("AP")] = Dictionary()
        self["AP"][Name("N")] = Stream()
        self["AP"]["N"][Name("Type")] = Name("XObject")
        self["AP"]["N"][Name("Subtype")] = Name("Form")

        appearance_stream_content = "q %f %f %f RG %f w 0 0 m " % (
            stroke_color.to_rgb().red,
            stroke_color.to_rgb().green,
            stroke_color.to_rgb().blue,
            stroke_width,
        )
        for x in range(0, int(bounding_box.width), 5):
            appearance_stream_content += "%f %f l %f %f l " % (x, 0, x + 2.5, 7)
        appearance_stream_content += "%f %f l " % (
            bounding_box.width - bounding_box.width % 5 + 5,
            0,
        )
        appearance_stream_content += "S Q"
        self["AP"]["N"][Name("DecodedBytes")] = bytes(
            appearance_stream_content, "latin1"
        )
        self["AP"]["N"][Name("Bytes")] = zlib.compress(
            self["AP"]["N"][Name("DecodedBytes")]
        )
        self["AP"]["N"][Name("Length")] = bDecimal(len(self["AP"]["N"][Name("Bytes")]))
        self["AP"]["N"][Name("Filter")] = Name("FlateDecode")

        # The lower-left corner of the bounding box (BBox) is set to coordinates (0, 0) in the form coordinate system.
        # The box’s top and right coordinates are taken from the dimensions of the annotation rectangle (the Rect
        # entry in the widget annotation dictionary).
        self["AP"]["N"][Name("BBox")] = List().set_is_inline(True)  # type: ignore [attr-defined]
        self["AP"]["N"]["BBox"].append(bDecimal(0))
        self["AP"]["N"]["BBox"].append(bDecimal(0))
        self["AP"]["N"]["BBox"].append(bDecimal(bounding_box.width))
        self["AP"]["N"]["BBox"].append(bDecimal(100))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This implementation of WriteBaseTransformer is responsible for writing /Catalog Dictionary objects
"""
import logging
import zlib
from pathlib import Path
from typing import Optional

from borb.io.read.types import (
    AnyPDFType,
    Dictionary,
    Name,
    String,
    List as bList,
    Stream,
    Decimal as bDecimal,
)
from borb.io.write.object.dictionary_transformer import DictionaryTransformer
from borb.io.write.transformer import WriteTransformerState

logger = logging.getLogger(__name__)


class CatalogTransformer(DictionaryTransformer):
    """
    This implementation of WriteBaseTransformer is responsible for writing /Catalog Dictionary objects
    """

    def can_be_transformed(self, any: AnyPDFType):
        """
        This function returns True if the object to be transformed is a /Catalog Dictionary
        """
        return (
            isinstance(any, Dictionary) and "Type" in any and any["Type"] == "Catalog"
        )

    #
    # /OutputIntents
    #

    def _build_rgb_outputintent_dictionary(self, root_dictionary: dict) -> None:

        # TODO check if exists already

        # read color profile bytes
        color_profile_bytes: bytes = b""
        with open(Path(__file__).parent / "resources/sRGB_CS_profile.icm", "rb") as fh:
            color_profile_bytes = fh.read()

        # create dest_output_profile
        dest_output_profile: Stream = Stream()
        dest_output_profile[Name("Alternate")] = Name("DeviceRGB")
        dest_output_profile[Name("DecodedBytes")] = color_profile_bytes
        dest_output_profile[Name("Bytes")] = zlib.compress(color_profile_bytes, 9)
        dest_output_profile[Name("Filter")] = Name("FlateDecode")
        dest_output_profile[Name("Length")] = bDecimal(len(color_profile_bytes))
        dest_output_profile[Name("N")] = bDecimal(3)

        # create RGB OutputIntent
        rgb_outputintent: dict = Dictionary()
        rgb_outputintent[Name("Type")] = Name("OutputIntent")
        rgb_outputintent[Name("S")] = Name("GTS_PDFA1")
        rgb_outputintent[Name("OutputCondition")] = String("")
        rgb_outputintent[Name("OutputConditionIdentifier")] = String("Custom")
        rgb_outputintent[Name("Info")] = String("sRGB IEC61966-2.1")
        rgb_outputintent[Name("RegistryName")] = String("http://www.color.org")
        rgb_outputintent[Name("DestOutputProfile")] = dest_output_profile
        dest_output_profile.set_parent(rgb_outputintent)  # type: ignore [attr-defined]

        # creatte OutputIntents
        outputintents_array = bList()
        outputintents_array.append(rgb_outputintent)
        rgb_outputintent.set_parent(outputintents_array)  # type: ignore [attr-defined]

        # add to root_dictionary
        root_dictionary[Name("OutputIntents")] = outputintents_array
        outputintents_array.set_parent(root_dictionary)  # type: ignore [attr-defined]

    def transform(
        self,
        object_to_transform: AnyPDFType,
        context: Optional[WriteTransformerState] = None,
    ):
        """
        This method writes a /Catalog Dictionary to a byte stream
        """

        # /OutputIntents
        # fmt: off
        needs_outputintents: bool = (context is not None and context.conformance_level is not None)
        if needs_outputintents:
            self._build_rgb_outputintent_dictionary(object_to_transform)
        # fmt: on

        # call super
        return super(CatalogTransformer, self).transform(object_to_transform, context)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Show one or more text strings, allowing individual glyph positioning. Each
element of array shall be either a string or a number.
"""
import typing
from typing import List

from borb.io.read.types import (
    AnyPDFType,
    Name,
    String,
    HexadecimalString,
    Decimal as bDecimal,
)
from borb.io.write.font.copy_command_operator import CopyCommandOperator
from borb.pdf.canvas.font.font import Font
from borb.pdf.canvas.operator.text.show_text_with_glyph_positioning import (
    ShowTextWithGlyphPositioning,
)


class SubSetShowTextWithGlyphPositioning(CopyCommandOperator):
    """
    Show one or more text strings, allowing individual glyph positioning. Each
    element of array shall be either a string or a number. If the element is a
    string, this operator shall show the string. If it is a number, the operator
    shall adjust the text position by that amount; that is, it shall translate the
    text matrix, T . The number shall be expressed in thousandths of a unit
    mof text space (see 9.4.4, "Text Space Details"). This amount shall be
    subtracted from the current horizontal or vertical coordinate, depending
    on the writing mode. In the default coordinate system, a positive
    adjustment has the effect of moving the next glyph painted either to the
    left or down by the given amount. Figure 46 shows an example of the
    effect of passing offsets to TJ.
    """

    def __init__(
        self, old_fonts: typing.List[Font], new_fonts: typing.List[Font], s: bytearray
    ):
        super(SubSetShowTextWithGlyphPositioning, self).__init__(
            ShowTextWithGlyphPositioning(), s
        )
        self._old_fonts: typing.List[Font] = old_fonts
        self._new_fonts: typing.List[Font] = new_fonts
        self._s: bytearray = s

    def _to_hex(self, i: int) -> str:
        s: str = hex(int(i))[2:]
        while len(s) < 2:
            s = "0" + s
        return s

    def invoke(
        self,
        canvas_stream_processor: "CanvasStreamProcessor",  # type: ignore [name-defined]
        operands: typing.List[AnyPDFType] = [],
        event_listeners: typing.List["EventListener"] = [],  # type: ignore [name-defined]
    ) -> None:
        """
        Invoke the TJ operator
        """

        assert isinstance(operands[0], List), "Operand 0 of TJ must be a List"
        canvas = canvas_stream_processor.get_canvas()

        # handle Font being a Name (optimization)
        assert canvas.graphics_state.font is not None
        font_name: typing.Optional[Name] = None
        if isinstance(canvas.graphics_state.font, Name):
            # fmt: off
            font_name = canvas.graphics_state.font
            canvas.graphics_state.font = canvas_stream_processor.get_resource("Font", canvas.graphics_state.font)
            # fmt: on

        # we're only interested in the old_font
        if canvas.graphics_state.font not in self._old_fonts:
            return super(SubSetShowTextWithGlyphPositioning, self).invoke(
                canvas_stream_processor, operands, event_listeners
            )

        old_font: Font = canvas.graphics_state.font
        new_font: Font = self._new_fonts[self._old_fonts.index(old_font)]

        # modify the operands
        operands_out: typing.List[AnyPDFType] = []
        for i in range(0, len(operands[0])):
            obj = operands[0][i]

            # adjust
            if isinstance(obj, bDecimal):
                operands_out.append(obj)
                continue

            # display string
            if isinstance(obj, String):
                old_char: str = old_font.character_identifier_to_unicode(
                    int(str(obj), 16)
                )
                new_code: str = self._to_hex(
                    new_font.unicode_to_character_identifier(old_char)
                )
                operands_out.append(HexadecimalString(new_code))

        # delegate
        return super(SubSetShowTextWithGlyphPositioning, self).invoke(
            canvas_stream_processor, [operands_out], event_listeners
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Set the text rendering mode, T mode , to render, which shall be an integer.
Initial value: 0.
"""

import typing

from borb.io.read.types import AnyPDFType
from borb.pdf.canvas.operator.canvas_operator import CanvasOperator


class SetTextRenderingMode(CanvasOperator):
    """
    Set the text rendering mode, T mode , to render, which shall be an integer.
    Initial value: 0.
    """

    def __init__(self):
        super().__init__("Tr", 1)

    def invoke(
        self,
        canvas_stream_processor: "CanvasStreamProcessor",  # type: ignore [name-defined]
        operands: typing.List[AnyPDFType] = [],
        event_listeners: typing.List["EventListener"] = [],  # type: ignore [name-defined]
    ) -> None:
        """
        Invoke the Tr operator
        """
        # TODO
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
In computer programming, a block or code block is a lexical structure of source code which is grouped together.
Blocks consist of one or more declarations and statements. A programming language that permits the creation of blocks,
including blocks nested within other blocks, is called a block-structured programming language.
Blocks are fundamental to structured programming, where control structures are formed from blocks.

The function of blocks in programming is to enable groups of statements to be treated as if they were one statement,
and to narrow the lexical scope of objects such as variables,
procedures and functions declared in a block so that they do not conflict with those having the same name used elsewhere.
In a block-structured programming language, the objects named in outer blocks are visible inside inner blocks,
unless they are masked by an object declared with the same name.
"""
import typing
from decimal import Decimal

from borb.pdf.canvas.color.color import Color, HexColor
from borb.pdf.canvas.font.font import Font
from borb.pdf.canvas.layout.layout_element import Alignment
from borb.pdf.canvas.layout.text.paragraph import Paragraph

try:
    import black  # type: ignore[import]

    able_to_import_black = True
except ImportError:
    able_to_import_black = False


class CodeBlock(Paragraph):
    """
    In computer programming, a block or code block is a lexical structure of source code which is grouped together.
    Blocks consist of one or more declarations and statements. A programming language that permits the creation of blocks,
    including blocks nested within other blocks, is called a block-structured programming language.
    Blocks are fundamental to structured programming, where control structures are formed from blocks.

    The function of blocks in programming is to enable groups of statements to be treated as if they were one statement,
    and to narrow the lexical scope of objects such as variables,
    procedures and functions declared in a block so that they do not conflict with those having the same name used elsewhere.
    In a block-structured programming language, the objects named in outer blocks are visible inside inner blocks,
    unless they are masked by an object declared with the same name.
    """

    def __init__(
        self,
        text: str,
        font: typing.Union[Font, str] = "Courier",
        font_size: Decimal = Decimal(12),
        font_color: Color = HexColor("24292e"),
        horizontal_alignment: Alignment = Alignment.LEFT,
        vertical_alignment: Alignment = Alignment.TOP,
        border_top: bool = False,
        border_right: bool = False,
        border_bottom: bool = False,
        border_left: bool = False,
        border_radius_top_left: Decimal = Decimal(0),
        border_radius_top_right: Decimal = Decimal(0),
        border_radius_bottom_right: Decimal = Decimal(0),
        border_radius_bottom_left: Decimal = Decimal(0),
        border_color: Color = HexColor("000000"),
        border_width: Decimal = Decimal(1),
        padding_top: Decimal = Decimal(5),
        padding_right: Decimal = Decimal(5),
        padding_bottom: Decimal = Decimal(5),
        padding_left: Decimal = Decimal(5),
        background_color: typing.Optional[Color] = HexColor("f6f8fa"),
    ):
        # format string using black
        if able_to_import_black:
            text = black.format_str(text, mode=black.Mode())

        # call super
        super().__init__(
            text=text,
            font=font,
            font_size=font_size,
            font_color=font_color,
            horizontal_alignment=horizontal_alignment,
            vertical_alignment=vertical_alignment,
            border_top=border_top,
            border_right=border_right,
            border_bottom=border_bottom,
            border_left=border_left,
            border_radius_top_left=border_radius_top_left,
            border_radius_top_right=border_radius_top_right,
            border_radius_bottom_right=border_radius_bottom_right,
            border_radius_bottom_left=border_radius_bottom_left,
            border_color=border_color,
            border_width=border_width,
            padding_top=padding_top,
            padding_right=padding_right,
            padding_bottom=padding_bottom,
            padding_left=padding_left,
            background_color=background_color,
            respect_newlines_in_text=True,
            respect_spaces_in_text=True,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Stroke the path.
"""

import typing

from borb.io.read.types import AnyPDFType
from borb.pdf.canvas.event.line_render_event import LineRenderEvent
from borb.pdf.canvas.operator.canvas_operator import CanvasOperator


class StrokePath(CanvasOperator):
    """
    Stroke the path.
    """

    def __init__(self):
        super().__init__("S", 0)

    def invoke(
        self,
        canvas_stream_processor: "CanvasStreamProcessor",  # type: ignore [name-defined]
        operands: typing.List[AnyPDFType] = [],
        event_listeners: typing.List["EventListener"] = [],  # type: ignore [name-defined]
    ) -> None:
        """
        Invoke the S operator
        """

        # get graphic state
        canvas = canvas_stream_processor.get_canvas()
        gs = canvas.graphics_state

        # notify listeners
        for el in event_listeners:
            for l in gs.path:
                el._event_occurred(LineRenderEvent(gs, l))

        # clear path
        gs.path = []



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Move to the start of the next line, offset from the start of the current line by
(tx , ty). As a side effect, this operator shall set the leading parameter in
the text state. This operator shall have the same effect as this code:
−ty TL
tx ty Td
"""
import typing
from decimal import Decimal

from borb.io.read.types import AnyPDFType
from borb.io.read.types import Decimal as bDecimal
from borb.pdf.canvas.operator.canvas_operator import CanvasOperator


class MoveTextPositionSetLeading(CanvasOperator):
    """
    Move to the start of the next line, offset from the start of the current line by
    (tx , ty). As a side effect, this operator shall set the leading parameter in
    the text state. This operator shall have the same effect as this code:
    −ty TL
    tx ty Td
    """

    def __init__(self):
        super().__init__("TD", 2)

    def invoke(
        self,
        canvas_stream_processor: "CanvasStreamProcessor",  # type: ignore [name-defined]
        operands: typing.List[AnyPDFType] = [],
        event_listeners: typing.List["EventListener"] = [],  # type: ignore [name-defined]
    ) -> None:
        """
        Invoke the TD operator
        """
        assert isinstance(operands[0], Decimal), "Operand 0 of TD must be a Decimal"
        assert isinstance(operands[1], Decimal), "Operand 1 of TD must be a Decimal"

        set_text_leading_op: typing.Optional[
            CanvasOperator
        ] = canvas_stream_processor.get_operator("TL")
        assert (
            set_text_leading_op
        ), "Operand TL must be defined for operator TD to function"
        set_text_leading_op.invoke(
            canvas_stream_processor, [bDecimal(-operands[1])], event_listeners
        )

        move_text_position_op: typing.Optional[
            CanvasOperator
        ] = canvas_stream_processor.get_operator("Td")
        assert (
            move_text_position_op
        ), "Operand Td must be defined for operator TD to function"
        move_text_position_op.invoke(canvas_stream_processor, operands, event_listeners)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    This module contains all classes needed to apply OCR (using Tesseract) to a PDF document.
"""
import logging
import typing
from decimal import Decimal
from pathlib import Path

from PIL import Image as PILImage  # type: ignore [import]
from PIL import ImageDraw

from borb.pdf.canvas.color.color import Color, HexColor, RGBColor
from borb.pdf.canvas.event.begin_page_event import BeginPageEvent
from borb.pdf.canvas.event.end_page_event import EndPageEvent
from borb.pdf.canvas.event.event_listener import Event, EventListener
from borb.pdf.canvas.event.image_render_event import ImageRenderEvent
from borb.pdf.canvas.font.font import Font
from borb.pdf.canvas.font.simple_font.font_type_1 import StandardType1Font
from borb.pdf.canvas.geometry.rectangle import Rectangle
from borb.pdf.page.page import Page

try:
    import pytesseract  # type: ignore [import]
    from pytesseract import Output  # type: ignore [import]
except ImportError:
    assert False, "Unable to import pytesseract"

logger = logging.getLogger(__name__)


class OCREvent(Event):
    """
    This implementation of Event represents content that was recognized during the OCR process.
    """

    def __init__(
        self,
        text: str,
        font: Font,
        font_size: Decimal,
        font_color: Color,
        bounding_box: Rectangle,
        page: Page,
        confidence: Decimal,
    ):
        self._text: str = text
        self._font: Font = font
        self._font_size: Decimal = font_size
        self._font_color: Color = font_color
        self._bounding_box: Rectangle = bounding_box
        self._page: Page = page
        self._confidence: Decimal = confidence

    def get_text(self) -> str:
        """
        This function returns the text of this OCREvent
        """
        return self._text

    def get_font(self) -> Font:
        """
        This function returns the Font of this OCREvent.
        This will likely be an estimate of the actual Font used in the Image
        """
        return self._font

    def get_font_size(self) -> Decimal:
        """
        This function returns the font_size of this OCREvent
        """
        return self._font_size

    def get_font_color(self) -> Color:
        """
        This function returns the font_color of this OCREvent
        """
        return self._font_color

    def get_bounding_box(self) -> Rectangle:
        """
        This function returns the bounding box of this OCREvent
        """
        return self._bounding_box

    def get_page(self) -> Page:
        """
        This function returns the Page of this OCREvent
        """
        return self._page

    def get_confidence(self) -> Decimal:
        """
        This function returns the OCR confidence of this OCREvent
        """
        return self._confidence


class OCRImageRenderEventListener(EventListener):
    """
    This implementation of EventListener attempts to perform OCR on Image objects inside a PDF.
    If text has been found, OCRImageRenderEventListener will add optional content to ensure
    the PDF can now be searched for the recognized text.
    """

    def __init__(
        self, tesseract_data_dir: Path, minimal_confidence: Decimal = Decimal(0.75)
    ):
        # fmt: off
        assert tesseract_data_dir.exists(), "OCRImageRenderEventListener requires the tesseract data directory."
        assert tesseract_data_dir.is_dir(), "OCRImageRenderEventListener requires the tesseract data directory."
        # fmt: on

        self._tesseract_data_dir: Path = tesseract_data_dir
        self._minimum_confidence: Decimal = minimal_confidence
        self._helvetica: Font = StandardType1Font("Helvetica")
        self._page: typing.Optional[Page] = None

    def _event_occurred(self, event: Event) -> None:
        if isinstance(event, BeginPageEvent):
            self._page = event.get_page()
            return

        if isinstance(event, EndPageEvent):
            self._page = None
            return

        if isinstance(event, ImageRenderEvent):

            data = pytesseract.image_to_data(
                event.get_image(),
                lang="eng",
                config='--tessdata-dir "%s"' % str(self._tesseract_data_dir.absolute()),
                output_type=Output.DICT,
            )

            width_ratio: Decimal = event.get_width() / event.get_image().width  # type: ignore[attr-defined]
            height_ratio: Decimal = event.get_height() / event.get_image().height  # type: ignore[attr-defined]

            number_of_boxes: int = len(data["level"])
            for i in range(0, number_of_boxes):

                x: Decimal = Decimal(data["left"][i])
                # tesseract considers (LEFT, TOP) to be the origin
                # PDF prefers (LEFT, BOTTOM)
                # the following code fixes the mismatch
                y: Decimal = (
                    Decimal(event.get_image().height)  # type: ignore [attr-defined]
                    - Decimal(data["top"][i])
                    - Decimal(data["height"][i])
                )

                image_bounding_box: Rectangle = Rectangle(
                    x, y, Decimal(data["width"][i]), Decimal(data["height"][i])
                )

                # convert bounding box to Rectangle object
                pdf_bounding_box: Rectangle = Rectangle(
                    x * width_ratio + event.get_x(),
                    y * height_ratio + event.get_y(),
                    Decimal(data["width"][i]) * width_ratio,
                    Decimal(data["height"][i]) * height_ratio,
                )

                # get text in bounding box
                text_in_bounding_box: str = data["text"][i]
                if text_in_bounding_box.strip() == "":
                    continue

                # get confidence
                confidence: Decimal = Decimal(data["conf"][i])
                if confidence < self._minimum_confidence:
                    continue

                # delegate call
                assert self._page is not None
                font_size: Decimal = round(
                    self._get_font_size(
                        text_in_bounding_box, pdf_bounding_box.get_width()
                    ),
                    0,
                )
                font_color: RGBColor = self._get_font_color(
                    text_in_bounding_box, event.get_image(), image_bounding_box
                )
                self._ocr_text_occurred(
                    OCREvent(
                        text_in_bounding_box,
                        self._helvetica,
                        font_size,
                        font_color,
                        pdf_bounding_box,
                        self._page,
                        confidence,
                    )
                )

    def _get_text_width(self, font_size: Decimal, text: str):
        w: Decimal = Decimal(0)
        ZERO: Decimal = Decimal(0)
        for c in text:
            try:
                cid: typing.Optional[
                    int
                ] = self._helvetica.unicode_to_character_identifier(c)
                assert cid is not None
                w += (
                    (self._helvetica.get_width(cid) or ZERO)
                    * font_size
                    * Decimal(0.001)
                )
            except:
                pass
        return w

    def _get_font_size(self, text: str, bounding_box_width: Decimal) -> Decimal:
        """
        This function attempts to find the font_size that would best fit the given text in the given width
        :param text:                the text to fit
        :param bounding_box_width:  the bounding box (width) to fit
        """
        estimated_font_size_lowerbound: Decimal = Decimal(1)
        estimated_font_size_upperbound: Decimal = Decimal(1024)
        iteration_count: int = 0
        while (
            abs(estimated_font_size_upperbound - estimated_font_size_lowerbound) > 1
            and iteration_count < 11
        ):
            midpoint: Decimal = (
                estimated_font_size_upperbound + estimated_font_size_lowerbound
            ) / Decimal(2)
            midpoint = Decimal(int(midpoint))
            estimated_width: Decimal = self._get_text_width(midpoint, text)
            iteration_count += 1
            if estimated_width > bounding_box_width:
                estimated_font_size_upperbound = midpoint
                continue
            if estimated_width < bounding_box_width:
                estimated_font_size_lowerbound = midpoint
                continue
        return estimated_font_size_lowerbound

    def _get_font_color(
        self,
        text: str,
        image: PILImage,  # type: ignore[valid-type]
        image_bounding_box: Rectangle,
    ) -> RGBColor:

        # build empty PILImage
        text_image: PILImage = PILImage.new(  # type: ignore[valid-type]
            "RGB",
            (int(image_bounding_box.get_width()), int(image_bounding_box.get_height())),
            color=(255, 255, 255),
        )

        # write text
        # this can go wrong if the default font for drawing text
        # does not support one or more of the characters being drawn
        # in which case this code returns black
        try:
            text_image_draw = ImageDraw.Draw(text_image)
            text_image_draw.text((0, 0), text, fill=(0, 0, 0))
        except:
            logger.debug(
                "Unable to write '%s' in default PIL font. No metrics available to determine color. Defaulting to black."
                % text
            )
            return HexColor("000000")

        # count number of text pixels
        try:
            percentage_of_text_pixels: Decimal = Decimal(0)
            max_x: int = 0
            max_y: int = 0
            for i in range(0, text_image.width):  # type: ignore[attr-defined]
                for j in range(0, text_image.height):  # type: ignore[attr-defined]
                    if text_image.getpixel((i, j)) == (0, 0, 0):  # type: ignore[attr-defined]
                        percentage_of_text_pixels += Decimal(1)
                        max_x = max(max_x, i)
                        max_y = max(max_y, j)
            percentage_of_text_pixels /= Decimal(max_x * max_y)
        except:
            logger.debug(
                "Unable to obtain metrics to determine color. Defaulting to black."
            )
            return HexColor("000000")

        # crop image
        cropped_image = image.crop(  # type: ignore[attr-defined]
            (
                image_bounding_box.x,
                image.height - image_bounding_box.y - image_bounding_box.height,  # type: ignore[attr-defined]
                image_bounding_box.x + image_bounding_box.width,  # type: ignore[attr-defined]
                image.height - image_bounding_box.y,  # type: ignore[attr-defined]
            )
        )

        # count number of pixels in cropped image
        number_of_pixels_in_cropped_image: Decimal = (
            cropped_image.width * cropped_image.height
        )

        # build color histogram
        color_histogram: typing.Dict[str, Decimal] = {}
        for i in range(0, cropped_image.width):
            for j in range(0, cropped_image.height):
                color_tuple: typing.Tuple[int, int, int] = cropped_image.getpixel(
                    (i, j)
                )
                color_tuple = (
                    color_tuple[0] - color_tuple[0] % 16,
                    color_tuple[1] - color_tuple[1] % 16,
                    color_tuple[2] - color_tuple[2] % 16,
                )
                hex_color: str = RGBColor(
                    Decimal(color_tuple[0] / 255),
                    Decimal(color_tuple[1] / 255),
                    Decimal(color_tuple[2] / 255),
                ).to_hex_string()
                color_histogram[hex_color] = color_histogram.get(
                    hex_color, Decimal(0)
                ) + Decimal(1)

        # normalize
        color_histogram = {
            k: (v / number_of_pixels_in_cropped_image)
            for k, v in color_histogram.items()
        }

        # trim
        color_histogram = {k: v for k, v in color_histogram.items() if v > 0.05}

        # find best match
        min_delta: typing.Optional[Decimal] = None
        min_delta_color: typing.Optional[RGBColor] = None
        for k, v in color_histogram.items():
            delta: Decimal = abs(percentage_of_text_pixels - v)
            if min_delta is None or delta < min_delta:
                min_delta = delta
                min_delta_color = HexColor(k)

        # return
        return min_delta_color or HexColor("000000")

    def _ocr_text_occurred(self, event: OCREvent):
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This class represents a generic disjoint shape (specified by a List of lines),
that will be colored according to a gradient.
It has convenience methods to calculate width and height, perform scaling, etc
"""
import enum
import math
from decimal import Decimal

from borb.pdf.canvas.color.color import Color, HSVColor, RGBColor
from borb.pdf.canvas.geometry.rectangle import Rectangle
from borb.pdf.canvas.layout.shape.disconnected_shape import DisconnectedShape
from borb.pdf.page.page import Page


class GradientColoredDisconnectedShape(DisconnectedShape):
    """
    This class represents a generic disconnected shape (specified by a List of lines),
    that will be colored according to a gradient.
    It has convenience methods to calculate width and height, perform scaling, etc
    """

    class GradientType(enum.Enum):
        """
        This enumeration represents the type
        of gradient that will be applied to this DisjointShape.
        """

        DIAGONAL = 1
        HORIZONTAL = 2
        RADIAL = 3
        VERTICAL = 4

    def __init__(
        self,
        shape: DisconnectedShape,
        from_color: Color,
        to_color: Color,
        gradient_type: GradientType = GradientType.RADIAL,
    ):
        super(GradientColoredDisconnectedShape, self).__init__(
            lines=shape._lines,
            stroke_color=shape._stroke_color,
            line_width=shape._line_width,
            background_color=shape._background_color,
            border_bottom=shape._border_bottom,
            border_color=shape._border_color,
            border_left=shape._border_left,
            border_radius_bottom_left=shape._border_radius_bottom_left,
            border_radius_bottom_right=shape._border_radius_bottom_right,
            border_radius_top_left=shape._border_radius_top_left,
            border_radius_top_right=shape._border_radius_top_right,
            border_right=shape._border_right,
            border_top=shape._border_top,
            border_width=shape._border_width,
            horizontal_alignment=shape._horizontal_alignment,
            margin_bottom=shape._margin_bottom,
            margin_left=shape._margin_left,
            margin_right=shape._margin_right,
            margin_top=shape._margin_top,
            padding_bottom=shape._padding_bottom,
            padding_left=shape._padding_left,
            padding_right=shape._padding_right,
            padding_top=shape._padding_top,
            vertical_alignment=shape._vertical_alignment,
        )
        self._from_color: Color = from_color
        self._to_color: Color = to_color
        self._gradient_type: GradientColoredDisconnectedShape.GradientType = (
            gradient_type
        )

    def _get_content_box(self, available_space: Rectangle) -> Rectangle:
        return super(GradientColoredDisconnectedShape, self)._get_content_box(
            available_space
        )

    def _paint_content_box(self, page: Page, bounding_box: Rectangle) -> None:

        # scale to fit
        self.scale_to_fit(bounding_box.width, bounding_box.height)

        # translate points to fit in box
        self.move_to(
            bounding_box.x, bounding_box.y + bounding_box.height - self.get_height()
        )

        # write content
        content = "q %d w " % (float(self._line_width),)
        min_x: Decimal = min([min(l[0][0], l[1][0]) for l in self._lines])
        min_y: Decimal = min([min(l[0][1], l[1][1]) for l in self._lines])
        max_x: Decimal = max([min(l[0][0], l[1][0]) for l in self._lines])
        max_y: Decimal = max([max(l[0][1], l[1][1]) for l in self._lines])

        # we need to calculate the norm (the max distance between a line and the origin of the gradient)
        n: Decimal = Decimal(1)

        # DIAGONAL
        if (
            self._gradient_type
            == GradientColoredDisconnectedShape.GradientType.DIAGONAL
        ):
            n = Decimal(math.sqrt((min_x - max_x) ** 2 + (min_y - max_y) ** 2))

        # HORIZONTAL
        if (
            self._gradient_type
            == GradientColoredDisconnectedShape.GradientType.HORIZONTAL
        ):
            n = max_x - min_x

        # RADIAL
        if self._gradient_type == GradientColoredDisconnectedShape.GradientType.RADIAL:
            mid_x = (max_x - min_x) / 2 + min_x
            mid_y = (max_y - min_y) / 2 + min_y
            n = Decimal(
                max(
                    [
                        math.sqrt((l[0][0] - mid_x) ** 2 + (l[0][1] - mid_y) ** 2)
                        for l in self._lines
                    ]
                )
            )

        # VERTICAL
        if (
            self._gradient_type
            == GradientColoredDisconnectedShape.GradientType.VERTICAL
        ):
            n = max_y - min_y

        # convert to Decimal to be sure
        n = Decimal(n)

        # calculate colors as HSV
        start_color: HSVColor = HSVColor.from_rgb(self._from_color.to_rgb())
        end_color: HSVColor = HSVColor.from_rgb(self._to_color.to_rgb())

        # now we can draw each line
        for l in self._lines:
            d: Decimal = Decimal(0)
            # fmt: off
            if self._gradient_type == GradientColoredDisconnectedShape.GradientType.DIAGONAL:
                d = Decimal(math.sqrt((l[0][0] - min_x) ** 2 + (l[0][1] - min_y) ** 2))
            if self._gradient_type == GradientColoredDisconnectedShape.GradientType.HORIZONTAL:
                d = Decimal(l[0][0] - min_x)
            if self._gradient_type == GradientColoredDisconnectedShape.GradientType.RADIAL:
                d = Decimal(math.sqrt((l[0][0] - mid_x) ** 2 + (l[0][1] - mid_y) ** 2))
            if self._gradient_type == GradientColoredDisconnectedShape.GradientType.VERTICAL:
                d = Decimal(l[0][1] - min_y)
            # fmt: on

            # calculate color based on the distance between this line and the origin of the gradient
            # fmt: off
            h = start_color.hue + (end_color.hue - start_color.hue) * (d / n)
            s = start_color.saturation + (end_color.saturation - start_color.saturation) * (d / n)
            v = start_color.value + (end_color.value - start_color.value) * (d / n)
            # fmt: on

            # convert to RGB
            stroke_color: RGBColor = HSVColor(h, s, v).to_rgb()
            r: float = float(stroke_color.red)
            g: float = float(stroke_color.green)
            b: float = float(stroke_color.blue)

            # write content
            content += "%f %f %f RG %f %f m %f %f l S " % (
                r,
                g,
                b,
                float(l[0][0]),
                float(l[0][1]),
                float(l[1][0]),
                float(l[1][1]),
            )

        # stroke
        content += " Q"

        # append to page
        page._append_to_content_stream(content)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This CanvasOperator copies an existing operator and writes its bytes to a content stream of the canvas.
"""
import typing
from decimal import Decimal

from borb.io.read.types import AnyPDFType, HexadecimalString, String, Name
from borb.pdf.canvas.operator.canvas_operator import CanvasOperator


class CopyCommandOperator(CanvasOperator):
    """
    This CanvasOperator copies an existing operator and writes its bytes to a content stream of the canvas.
    """

    def __init__(
        self, operator_to_copy: CanvasOperator, output_content_stream: bytearray
    ):
        super().__init__("", 0)
        self._operator_to_copy = operator_to_copy
        self._output_content_stream: bytearray = output_content_stream

    def get_text(self) -> str:
        """
        Return the str that invokes this CanvasOperator
        """
        return self._operator_to_copy.get_text()

    def get_number_of_operands(self) -> int:
        """
        Return the number of operands for this CanvasOperator
        """
        return self._operator_to_copy.get_number_of_operands()

    def _operand_to_str(self, op: AnyPDFType) -> str:
        if isinstance(op, Decimal):
            return str(op)
        if isinstance(op, HexadecimalString):
            return "<" + op._text + ">"
        if isinstance(op, String):
            return "(" + op._text + ")"
        if isinstance(op, Name):
            return "/" + str(op)
        if isinstance(op, list):
            return "[" + "".join([self._operand_to_str(x) + " " for x in op])[:-1] + "]"
        return ""

    def invoke(
        self,
        canvas_stream_processor: "CanvasStreamProcessor",  # type: ignore [name-defined]
        operands: typing.List[AnyPDFType] = [],
        event_listeners: typing.List["EventListener"] = [],  # type: ignore [name-defined]
    ) -> None:
        """
        Invokes this CanvasOperator
        """

        # execute command
        self._operator_to_copy.invoke(canvas_stream_processor, operands)

        # copy command in content stream
        canvas = canvas_stream_processor.get_canvas()

        # copy operand string
        self._output_content_stream += b"\n"
        self._output_content_stream += b"".join(
            [(bytes(self._operand_to_str(s), encoding="utf8") + b" ") for s in operands]
        )
        self._output_content_stream += bytes(self.get_text(), encoding="utf8")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    This Enum provides a convenient way of getting all common paper page sizes
"""
import enum
from decimal import Decimal


class PageSize(enum.Enum):
    """
    This Enum provides a convenient way of getting all common paper page sizes
    """

    A0_PORTRAIT = (Decimal(2384), Decimal(3370))
    A0_LANDSCAPE = (Decimal(3370), Decimal(2384))

    A1_PORTRAIT = (Decimal(1684), Decimal(2384))
    A1_LANDSCAPE = (Decimal(2384), Decimal(1684))

    A2_PORTRAIT = (Decimal(1190), Decimal(1684))
    A2_LANDSCAPE = (Decimal(1684), Decimal(1190))

    A3_PORTRAIT = (Decimal(842), Decimal(1190))
    A3_LANDSCAPE = (Decimal(842), Decimal(1190))

    A4_PORTRAIT = (Decimal(595), Decimal(842))
    A4_LANDSCAPE = (Decimal(842), Decimal(595))

    A5_PORTRAIT = (Decimal(420), Decimal(595))
    A5_LANDSCAPE = (Decimal(595), Decimal(420))

    A6_PORTRAIT = (Decimal(298), Decimal(420))
    A6_LANDSCAPE = (Decimal(420), Decimal(298))

    A7_PORTRAIT = (Decimal(210), Decimal(298))
    A7_LANDSCAPE = (Decimal(298), Decimal(210))

    A8_PORTRAIT = (Decimal(148), Decimal(210))
    A8_LANDSCAPE = (Decimal(210), Decimal(148))

    A9_PORTRAIT = (Decimal(105), Decimal(547))
    A9_LANDSCAPE = (Decimal(547), Decimal(105))

    A10_PORTRAIT = (Decimal(74), Decimal(105))
    A10_LANDSCAPE = (Decimal(105), Decimal(74))

    B0_PORTRAIT = (Decimal(2834), Decimal(4008))
    B0_LANDSCAPE = (Decimal(4008), Decimal(2834))

    B1_PORTRAIT = (Decimal(2004), Decimal(2834))
    B1_LANDSCAPE = (Decimal(2834), Decimal(2004))

    B2_PORTRAIT = (Decimal(1417), Decimal(2004))
    B2_LANDSCAPE = (Decimal(2004), Decimal(1417))

    B3_PORTRAIT = (Decimal(1000), Decimal(1417))
    B3_LANDSCAPE = (Decimal(1417), Decimal(1000))

    B4_PORTRAIT = (Decimal(708), Decimal(1000))
    B4_LANDSCAPE = (Decimal(1000), Decimal(708))

    B5_PORTRAIT = (Decimal(498), Decimal(708))
    B5_LANDSCAPE = (Decimal(708), Decimal(498))

    B6_PORTRAIT = (Decimal(354), Decimal(498))
    B6_LANDSCAPE = (Decimal(498), Decimal(354))

    B7_PORTRAIT = (Decimal(249), Decimal(354))
    B7_LANDSCAPE = (Decimal(354), Decimal(249))

    B8_PORTRAIT = (Decimal(175), Decimal(249))
    B8_LANDSCAPE = (Decimal(249), Decimal(175))

    B9_PORTRAIT = (Decimal(124), Decimal(175))
    B9_LANDSCAPE = (Decimal(175), Decimal(124))

    B10_PORTRAIT = (Decimal(88), Decimal(124))
    B10_LANDSCAPE = (Decimal(124), Decimal(88))

    LETTER_PORTRAIT = (Decimal(612), Decimal(792))
    LETTER_LANDSCAPE = (Decimal(792), Decimal(612))

    LEGAL_PORTRAIT = (Decimal(612), Decimal(1008))
    LEGAL_LANDSCAPE = (Decimal(1008), Decimal(612))

    TABLOID_PORTRAIT = (Decimal(792), Decimal(1224))
    TABLOID_LANDSCAPE = (Decimal(792), Decimal(1224))

    LEDGER_PORTRAIT = (Decimal(1224), Decimal(792))
    LEDGER_LANDSCAPE = (Decimal(792), Decimal(1224))

    EXECUTIVE_PORTRAIT = (Decimal(522), Decimal(756))
    EXECUTIVE_LANDSCAPE = (Decimal(756), Decimal(522))



# !/usr/bin/env python
# -*- coding: utf-8 -*-
"""
One commonly used font encoding for Latin-text font programs is often referred to as StandardEncoding or
sometimes as the Adobe standard encoding. The name StandardEncoding shall have no special meaning in
PDF, but this encoding does play a role as a default encoding (as shown in Table 114).

This is the built-in encoding defined in Type 1 Latin-text font programs (but generally not in TrueType font
programs). Conforming readers shall not have a predefined encoding named StandardEncoding. However, it is necessary to describe this
encoding, since a font’s built-in encoding can be used as the base encoding from which differences may be specified in an encoding
dictionary.
"""

# fmt: off
ADOBE_STANDARD_ENCODING_LOOKUP = [
    0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0,
    32, 33, 34, 35, 36, 37, 38, 39,
    40, 41, 42, 43, 44, 45, 46, 47,
    48, 49, 50, 51, 52, 53, 54, 55,
    56, 57, 58, 59, 60, 61, 62, 63,
    64, 65, 66, 67, 68, 69, 70, 71,
    72, 73, 74, 75, 76, 77, 78, 79,
    80, 81, 82, 83, 84, 85, 86, 87,
    88, 89, 90, 91, 92, 93, 94, 95,
    8216, 97, 98, 99, 100, 101, 102, 103,
    104, 105, 106, 107, 108, 109, 110, 111,
    112, 113, 114, 115, 116, 117, 118, 119,
    120, 121, 122, 123, 124, 125, 126, 0,
    0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0,
    0, 161, 162, 163, 8260, 165, 402, 167,
    164, 39, 8220, 171, 8249, 8250, 64257, 64258,
    0, 8211, 8224, 8225, 183, 0, 182, 8226,
    8218, 8222, 8221, 187, 8230, 8240, 0, 191,
    0, 96, 180, 710, 732, 175, 728, 729,
    168, 0, 730, 184, 0, 733, 731, 711,
    8212, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0,
    0, 198, 0, 170, 0, 0, 0, 0,
    321, 216, 338, 186, 0, 0, 0, 0,
    0, 230, 0, 0, 0, 305, 0, 0,
    322, 248, 339, 223, 0, 0, 0, 0,
]
# fmt: on


def adobe_standard_encode(str_input: str) -> bytes:
    """
    This function encodes a str using StandardEncoding
    """
    b: bytearray = bytearray()
    for c in str_input:
        char_index: int = -1
        try:
            char_index = ADOBE_STANDARD_ENCODING_LOOKUP.index(ord(c))
        except ValueError:
            pass
        if char_index != -1:
            b.append(char_index)
    return b


def adobe_standard_decode(byte_input: bytes) -> str:
    """
    This function decodes bytes using StandardEncoding
    """
    s: str = ""
    for b in byte_input:
        s += chr(ADOBE_STANDARD_ENCODING_LOOKUP[b])
    return s



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    This implementation of EventListener  extracts the colors used in rendering a PDF
"""
import typing
from decimal import Decimal
from typing import Optional

from PIL.Image import Image  # type: ignore [import]

from borb.pdf.canvas.color.color import RGBColor
from borb.pdf.canvas.event.begin_page_event import BeginPageEvent
from borb.pdf.canvas.event.chunk_of_text_render_event import ChunkOfTextRenderEvent
from borb.pdf.canvas.event.event_listener import Event, EventListener
from borb.pdf.canvas.event.image_render_event import ImageRenderEvent
from borb.pdf.canvas.geometry.rectangle import Rectangle
from borb.pdf.page.page import Page


class ColorSpectrumExtraction(EventListener):
    """
    This implementation of EventListener  extracts the colors used in rendering a PDF
    """

    def __init__(self, maximum_number_of_colors: Optional[int] = None):
        """
        Constructs a new ColorSpectrumExtraction
        """
        self._maximum_number_of_colors = 64
        if maximum_number_of_colors is not None:
            self._maximum_number_of_colors = maximum_number_of_colors
        self._colors_per_page: typing.Dict[
            int, typing.Dict[typing.Tuple[int, int, int], Decimal]
        ] = {}
        self._current_page: int = -1

    def _event_occurred(self, event: Event) -> None:
        if isinstance(event, BeginPageEvent):
            self._begin_page(event.get_page())
        if isinstance(event, ChunkOfTextRenderEvent):
            self._render_text(event)
        if isinstance(event, ImageRenderEvent):
            self._render_image(event)

    def _begin_page(self, page: Page):
        self._current_page += 1
        self._colors_per_page[self._current_page] = {}

    def _render_text(self, event: ChunkOfTextRenderEvent):
        assert event is not None
        bb: typing.Optional[Rectangle] = event.get_previous_layout_box()
        s: Decimal = Decimal(0) if bb is None else (bb.width * bb.height)
        c: RGBColor = event._font_color.to_rgb()
        self._register_color(s, c)

    def _render_image(self, event: ImageRenderEvent):
        r = (event.get_width() * event.get_height()) / (event.get_image().width * event.get_image().height)  # type: ignore[attr-defined]
        color_count: typing.Dict[RGBColor, Decimal] = {}
        for i in range(0, event.get_image().width):  # type: ignore [attr-defined]
            for j in range(0, event.get_image().height):  # type: ignore [attr-defined]
                c = ColorSpectrumExtraction._get_rgb_from_image(event.get_image(), i, j)
                if c not in color_count:
                    color_count[c] = Decimal(1)
                else:
                    color_count[c] += Decimal(1)
        for k, v in color_count.items():
            self._register_color(v * r, k)

    @staticmethod
    def _get_rgb_from_image(img: Image, x: int, y: int) -> RGBColor:
        c = img.getpixel((x, y))
        if img.mode == "RGB":
            return RGBColor(
                r=Decimal(c[0] / 255),
                g=Decimal(c[1] / 255),
                b=Decimal(c[2] / 255),
            )
        if img.mode == "RGBA":
            return RGBColor(
                r=Decimal(c[0] / 255),
                g=Decimal(c[1] / 255),
                b=Decimal(c[2] / 255),
            )
        if img.mode == "CMYK":
            r = int((1 - c[0]) * (1 - c[3]))
            g = int((1 - c[1]) * (1 - c[3]))
            b = int((1 - c[2]) * (1 - c[0]))
            return RGBColor(r=Decimal(r), g=Decimal(g), b=Decimal(b))
        return RGBColor(Decimal(0), Decimal(0), Decimal(0))

    def _register_color(self, amount: Decimal, color: RGBColor):
        mod_step = int(255 / (self._maximum_number_of_colors ** (1.0 / 3)))
        r = int(color.to_rgb().red * 255)
        r = r - r % mod_step

        g = int(color.to_rgb().green * 255)
        g = g - g % mod_step

        b = int(color.to_rgb().blue * 255)
        b = b - b % mod_step

        t = (r, g, b)
        if t not in self._colors_per_page[self._current_page]:
            self._colors_per_page[self._current_page][t] = amount
        else:
            self._colors_per_page[self._current_page][t] += amount

    def get_colors_for_page(
        self, page_number: int, limit: Optional[int] = None
    ) -> typing.List[typing.Tuple[RGBColor, Decimal]]:
        """
        This function returns the colors used on a given page of a PDF
        """
        if limit is None:
            limit = 32
        tmp = sorted(
            [
                (
                    RGBColor(
                        Decimal(k[0] / 255), Decimal(k[1] / 255), Decimal(k[2] / 255)
                    ),
                    v,
                )
                for k, v in self._colors_per_page[page_number].items()
            ],
            key=lambda x: x[1],
            reverse=True,
        )[0:limit]
        return tmp



# !/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Base Transformer implementation.
Add children to handle specific cases (transforming dictionaries, arrays, xref, etc)
"""
import io
import typing
from typing import Any, Optional, Union

from borb.io.read.tokenize.high_level_tokenizer import HighLevelTokenizer
from borb.io.read.types import AnyPDFType, Reference
from borb.pdf.canvas.event.event_listener import EventListener


class ReadTransformerState:
    """
    This class represents all the meta-information used in the process of reading a PDF document.
    This includes:
    - the root object (the Document itself)
    - the tokenizer
    - references that have been resolved (to avoid endless loops)
    - etc
    """

    def __init__(
        self,
        source: Optional[Union[io.BufferedIOBase, io.RawIOBase, io.BytesIO]] = None,
        tokenizer: Optional[HighLevelTokenizer] = None,
        root_object: Optional[Any] = None,
        password: typing.Optional[str] = None,
    ):
        self.source = source
        self.tokenizer = tokenizer
        self.root_object = root_object
        self.indirect_reference_chain: typing.Set[Reference] = set()
        self.password: typing.Optional[str] = password
        self.security_handler: typing.Optional[typing.Any] = None


class Transformer:
    """
    Base Transformer implementation.
    Add children to handle specific cases (transforming dictionaries, arrays, xref, etc)
    """

    def __init__(self):
        self._children = []
        self._parent = None
        self._level = 0
        self._invocation_count = 0

    def get_children(self) -> typing.List["Transformer"]:
        """
        This functions returns all child ReadBaseTransformers of this ReadBaseTransformer
        """
        return self._children

    def add_child_transformer(self, child_transformer: "Transformer") -> "Transformer":  # type: ignore[name-defined]
        """
        Add a child ReadBaseTransformer to this ReadBaseTransformer.
        Child transformers can be used to encapsulate specific object-creation/transformation logic.
        e.g. creating XREF, converting arrays, dictionaries, etc
        :param handler: the ReadBaseTransformer implementation to be added
        :type handler:  Transformer
        """
        self._children.append(child_transformer)
        child_transformer._parent = self
        return self

    def get_root_transformer(self) -> "Transformer":  # type: ignore[name-defined]
        """
        This function returns the top-level ReadBaseTransformer.
        This is useful when delegating calls from inside a ReadBaseTransformer to another.
        e.g. Transforming a dictionary by reading its delimiters and then transforming each key/value in turn by
        calling top-level ReadBaseTransformer
        """
        p = self
        while p._parent is not None:
            p = p._parent
        return p

    def can_be_transformed(
        self, object: Union[io.BufferedIOBase, io.RawIOBase, io.BytesIO, AnyPDFType]
    ) -> bool:
        """
        This function returns True if the object to be transformed can be transformed by this ReadBaseTransformer
        """
        return False

    def _is_recursive_object(self, object_to_transform) -> bool:
        try:
            p = object_to_transform
            parents = [p]
            while p is not None:
                p = p.get_parent()
                if p in parents:
                    return True
                parents.append(p)
        except:
            pass
        return False

    def transform(
        self,
        object_to_transform: Union[io.BufferedIOBase, io.RawIOBase, AnyPDFType],
        parent_object: Any,
        context: Optional[ReadTransformerState] = None,
        event_listeners: typing.List[EventListener] = [],
    ) -> Any:
        """
        This function reads an object from a byte stream.
        The object being read depends on the implementation of ReadBaseTransformer.
        """
        if self._is_recursive_object(object_to_transform):
            return object_to_transform
        for h in self._children:
            if h.can_be_transformed(object_to_transform):
                # print("%s<%s level='%d' invocation='%d'>" % ("   " * self.level, h.__class__.__name__, self.level, self.invocation_count), flush=True)
                self._level += 1
                self._invocation_count += 1
                out = h.transform(
                    object_to_transform,
                    parent_object=parent_object,
                    context=context,
                    event_listeners=event_listeners,
                )
                self._level -= 1
                # print("%s</%s>" % ("   " * self.level, h.__class__.__name__))
                return out
        return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    This implementation of Event is triggered right after the Canvas has processed a text-rendering instruction
"""
import typing
from decimal import Decimal

from borb.io.read.types import String
from borb.pdf.canvas.canvas_graphics_state import CanvasGraphicsState
from borb.pdf.canvas.event.event_listener import Event
from borb.pdf.canvas.font.font import Font
from borb.pdf.canvas.font.glyph_line import GlyphLine
from borb.pdf.canvas.geometry.rectangle import Rectangle
from borb.pdf.canvas.layout.text.chunk_of_text import ChunkOfText


class ChunkOfTextRenderEvent(Event, ChunkOfText):
    """
    This implementation of Event is triggered right after the Canvas has processed a text-rendering instruction
    """

    def __init__(self, graphics_state: CanvasGraphicsState, raw_bytes: String):
        assert graphics_state.font is not None
        assert isinstance(graphics_state.font, Font)
        self._glyph_line: GlyphLine = GlyphLine.from_bytes(
            raw_bytes.get_value_bytes(),
            graphics_state.font,
            graphics_state.font_size,
            graphics_state.character_spacing,
            graphics_state.word_spacing,
            graphics_state.horizontal_scaling,
        )
        super(ChunkOfTextRenderEvent, self).__init__(
            font=graphics_state.font,
            font_size=graphics_state.font_size * graphics_state.text_matrix[1][1],
            font_color=graphics_state.non_stroke_color,
            text=self._glyph_line.get_text(),
        )
        m = graphics_state.text_matrix.mul(graphics_state.ctm)
        m[1][1] *= graphics_state.font_size

        # calculate baseline box
        p0 = m.cross(Decimal(0), graphics_state.text_rise, Decimal(1))
        p1 = m.cross(
            self._glyph_line.get_width_in_text_space(),
            graphics_state.text_rise
            + graphics_state.font.get_ascent() * Decimal(0.001),
            Decimal(1),
        )

        # set baseline box
        self._baseline_bounding_box = Rectangle(
            min(p0[0], p1[0]), min(p0[1], p1[1]), abs(p1[0] - p0[0]), abs(p1[1] - p0[1])
        )

        # calculate bounding box
        uses_descent = any(
            [x in self._text.lower() for x in ["y", "p", "q", "f", "g", "j"]]
        )
        if uses_descent:
            p0 = m.cross(
                Decimal(0),
                graphics_state.text_rise
                + graphics_state.font.get_descent() * Decimal(0.001),
                Decimal(1),
            )
            p1 = m.cross(
                self._glyph_line.get_width_in_text_space(),
                graphics_state.text_rise
                + graphics_state.font.get_ascent() * Decimal(0.001),
                Decimal(1),
            )
            self._previous_layout_box = Rectangle(
                min(p0[0], p1[0]),
                min(p0[1], p1[1]),
                abs(p1[0] - p0[0]),
                abs(p1[1] - p0[1]),
            )
        else:
            self._previous_layout_box = self._baseline_bounding_box

        # calculate space character width estimate
        current_font: Font = graphics_state.font
        self._space_character_width_estimate_in_user_space = (
            current_font.get_space_character_width_estimate()
            * graphics_state.font_size
            * graphics_state.text_matrix[0][0]
            * Decimal(0.001)
        )
        assert graphics_state.font_size is not None
        self._font_size: Decimal = (
            graphics_state.font_size * graphics_state.text_matrix[0][0]
        )

        # store graphics state
        self._graphics_state = graphics_state

    def get_font_size(self) -> Decimal:
        """
        This function returns the font size
        """
        return self._font_size

    def get_space_character_width_estimate_in_user_space(self) -> Decimal:
        """
        This function returns the width (in user space) of the space-character.
        """
        return self._space_character_width_estimate_in_user_space

    def get_space_character_width_estimate_in_text_space(self) -> Decimal:
        """
        This function returns the width (in text space) of the space-character.
        """
        return (
            self._space_character_width_estimate_in_user_space
            * Decimal(1000)
            / self._font_size
        )

    def get_baseline(self) -> Rectangle:
        """
        This function returns the bounding box of this ChunkOfTextRenderEvent,
        starting at the baseline (not at the descent)
        """
        return self._baseline_bounding_box

    def split_on_glyphs(self) -> typing.List["ChunkOfTextRenderEvent"]:
        """
        This function splits this ChunkOfTextRenderEvent on every Glyph
        """
        chunks_of_text: typing.List[ChunkOfTextRenderEvent] = []
        x: Decimal = Decimal(0)
        y: Decimal = self._graphics_state.text_rise
        assert isinstance(self._graphics_state.font, Font)
        assert self._graphics_state.font is not None
        font: Font = self._graphics_state.font
        for g in self._glyph_line.split():
            e = ChunkOfTextRenderEvent(self._graphics_state, String(" "))
            e._font_size = self._font_size
            e._font_color = self._font_color
            e._font = self._font
            e._text = g.get_text()
            e._space_character_width_estimate_in_user_space = (
                self._space_character_width_estimate_in_user_space
            )
            e._graphics_state = self._graphics_state
            e._glyph_line = g

            # set baseline bounding box
            m = self._graphics_state.text_matrix.mul(self._graphics_state.ctm)
            m[1][1] *= self._graphics_state.font_size
            p0 = m.cross(x, y, Decimal(1))
            p1 = m.cross(
                x + g.get_width_in_text_space(),
                y + font.get_ascent() * Decimal(0.001),
                Decimal(1),
            )
            e._baseline_bounding_box = Rectangle(
                p0[0], p0[1], p1[0] - p0[0], p1[1] - p0[1]
            )
            e._previous_layout_box = e._baseline_bounding_box

            # change bounding box (descent)
            if g.uses_descent():
                p0 = m.cross(
                    x,
                    y + font.get_descent() * Decimal(0.001),
                    Decimal(1),
                )
                p1 = m.cross(
                    x + g.get_width_in_text_space(),
                    y + font.get_ascent() * Decimal(0.001),
                    Decimal(1),
                )
                e._previous_layout_box = Rectangle(
                    min(p0[0], p1[0]),
                    min(p0[1], p1[1]),
                    abs(p1[0] - p0[0]),
                    abs(p1[1] - p0[1]),
                )

            # update x
            x += g.get_width_in_text_space()

            # append
            chunks_of_text.append(e)

        return chunks_of_text


class LeftToRightComparator:
    """
    This class offers a comparator on ChunkOfTextRenderEvent objects.
    This comparator favors left-to-right, up-to-down text reading order.
    This corresponds to the expected western language reading order.
    """

    @staticmethod
    def cmp(obj0: ChunkOfTextRenderEvent, obj1: ChunkOfTextRenderEvent):
        """
        This function compares two ChunkOfTextRenderEvent objects
        returning a negative number if obj0 occurs first in the (western) reading order,
        and a positive number otherwise.
        """
        # get baseline
        y0_round = obj0.get_baseline().y
        y0_round = y0_round - y0_round % 5

        # get baseline
        y1_round = obj1.get_baseline().y
        y1_round = y1_round - y1_round % 5

        if y0_round == y1_round:
            return obj0.get_baseline().x - obj1.get_baseline().x
        return -(y0_round - y1_round)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This implementation of ReadBaseTransformer is responsible for reading a Dictionary object
"""
import io
import typing
from typing import Any, Optional, Union

from borb.io.read.transformer import ReadTransformerState, Transformer
from borb.io.read.types import AnyPDFType, Dictionary
from borb.pdf.canvas.event.event_listener import EventListener


class DictionaryTransformer(Transformer):
    """
    This implementation of ReadBaseTransformer is responsible for reading a Dictionary object
    """

    def can_be_transformed(
        self, object: Union[io.BufferedIOBase, io.RawIOBase, io.BytesIO, AnyPDFType]
    ) -> bool:
        """
        This function returns True if the object to be transformed is a Dictionary object
        """
        return isinstance(object, Dictionary)

    def transform(
        self,
        object_to_transform: Union[io.BufferedIOBase, io.RawIOBase, AnyPDFType],
        parent_object: Any,
        context: Optional[ReadTransformerState] = None,
        event_listeners: typing.List[EventListener] = [],
    ) -> Any:
        """
        This function reads a Dictionary from a byte stream
        """

        # create root object
        # fmt: off
        assert isinstance(object_to_transform, Dictionary), "object_to_transform must be of type Dictionary"
        # fmt: on
        object_to_transform.set_parent(parent_object)  # type: ignore [attr-defined]

        # transform key/value pair(s)
        for k, v in object_to_transform.items():
            v = self.get_root_transformer().transform(
                v, object_to_transform, context, event_listeners
            )
            if v is not None:
                object_to_transform[k] = v

        # return
        return object_to_transform



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Fill the path, using the nonzero winding number rule to determine the region
to fill (see 8.5.3.3.2, "Nonzero Winding Number Rule"). Any subpaths that
are open shall be implicitly closed before being filled.
"""
